from __future__ import print_function
from past.builtins import basestring
import os.path as op
import operator
from functools import reduce
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterface, isdefined)
from arcana.exceptions import ArcanaUsageError


class TestMathInputSpec(TraitedSpec):

    x = traits.Either(traits.Float(), traits.File(exists=True),
                      traits.List(traits.Float),
                      traits.List(traits.File(exists=True)),
                      desc='first arg')
    y = traits.Either(traits.Float(), traits.File(exists=True),
                      mandatory=False, desc='second arg')
    op = traits.Str(mandatory=True, desc='operation')

    z = traits.File(genfile=True, mandatory=False,
                    desc="Name for output file")

    as_file = traits.Bool(False, desc="Whether to write as a file",
                          usedefault=True)


class TestMathOutputSpec(TraitedSpec):

    z = traits.Either(traits.Float(), traits.File(exists=True),
                      'output')


class TestMath(BaseInterface):
    """
    A basic interface to test out the pipeline infrastructure
    """

    input_spec = TestMathInputSpec
    output_spec = TestMathOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        x = self.inputs.x
        y = self.inputs.y
        if isinstance(x, basestring):
            x = self._load_file(x)
        if isinstance(y, basestring):
            y = self._load_file(y)
        oper = getattr(operator, self.inputs.op)
        if isdefined(y):
            z = oper(x, y)
        elif isinstance(x, list):
            if x:
                if isinstance(x[0], basestring):
                    x = [self._load_file(u) for u in x]
            else:
                raise ArcanaUsageError("Cannot provide empty list to 'x'")
            z = reduce(oper, x)
        else:
            raise Exception(
                "If 'y' is not provided then x needs to be list")
        outputs = self.output_spec().get()
        if self.inputs.as_file:
            z_path = op.abspath(self._gen_z_fname())
            with open(z_path, 'w') as f:
                f.write(str(z))
            outputs['z'] = z_path
        else:
            outputs['z'] = z
        return outputs

    def _gen_filename(self, name):
        if name == 'z':
            fname = self._gen_z_fname()
        else:
            assert False
        return fname

    def _gen_z_fname(self):
        if isdefined(self.inputs.z):
            fname = self.inputs.z
        else:
            fname = 'z.txt'
        return fname

    @classmethod
    def _load_file(self, path):
        with open(path) as f:
            try:
                return float(f.read())
            except:
                raise



from past.builtins import basestring
import os
from itertools import chain
import os.path as op
import hashlib
from arcana.utils import split_extension, parse_value
from arcana.exceptions import (
    ArcanaError, ArcanaFileFormatError, ArcanaUsageError, ArcanaNameError,
    ArcanaDataNotDerivedYetError)
from .file_format import FileFormat
from .base import BaseFileset, BaseField


class BaseItem(object):

    is_spec = False

    def __init__(self, subject_id, visit_id, repository, from_study,
                 exists, record):
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._repository = repository
        self._from_study = from_study
        self._exists = exists
        self._record = record

    def __eq__(self, other):
        return (self.subject_id == other.subject_id and
                self.visit_id == other.visit_id and
                self.from_study == other.from_study and
                self.exists == other.exists and
                self._record == other._record)

    def __hash__(self):
        return (hash(self.subject_id) ^
                hash(self.visit_id) ^
                hash(self.from_study) ^
                hash(self.exists) ^
                hash(self._record))

    def find_mismatch(self, other, indent=''):
        sub_indent = indent + '  '
        mismatch = ''
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        if self.from_study != other.from_study:
            mismatch += ('\n{}from_study: self={} v other={}'
                         .format(sub_indent, self.from_study,
                                 other.from_study))
        if self.exists != other.exists:
            mismatch += ('\n{}exists: self={} v other={}'
                         .format(sub_indent, self.exists,
                                 other.exists))
        if self._record != other._record:
            mismatch += ('\n{}_record: self={} v other={}'
                         .format(sub_indent, self._record,
                                 other._record))
        return mismatch

    @property
    def derived(self):
        return self.from_study is not None

    @property
    def repository(self):
        return self._repository

    @property
    def exists(self):
        return self._exists

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def visit_id(self):
        return self._visit_id

    @property
    def session_id(self):
        return (self.subject_id, self.visit_id)

    @property
    def from_study(self):
        return self._from_study

    @property
    def record(self):
        return self._record

    @record.setter
    def record(self, record):
        if self.name not in record.outputs:
            raise ArcanaNameError(
                self.name,
                "{} was not found in outputs {} of provenance record".format(
                    self.name, record.outputs.keys(), record))
        self._record = record

    @property
    def recorded_checksums(self):
        if self.record is None:
            return None
        else:
            return self._record.outputs[self.name]

    def initkwargs(self):
        dct = super(Fileset, self).initkwargs()
        dct['repository'] = self.repository
        dct['subject_id'] = self.subject_id
        dct['visit_id'] = self.visit_id
        dct['from_study'] = self._from_study
        return dct


class Fileset(BaseItem, BaseFileset):
    """
    A representation of a fileset within the repository.

    Parameters
    ----------
    name : str
        The name of the fileset
    format : FileFormat
        The file format used to store the fileset.
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    derived : bool
        Whether the scan was generated or acquired. Depending on the repository
        used to store the fileset this is used to determine the location of the
        fileset.
    path : str | None
        The path to the fileset (for repositories on the local system)
    aux_files : dict[str, str] | None
        Additional files in the fileset. Keys should match corresponding
        aux_files dictionary in format.
    id : int | None
        The ID of the fileset in the session. To be used to
        distinguish multiple filesets with the same scan type in the
        same session, e.g. scans taken before and after a task. For
        repositorys where this isn't stored (i.e. Local), id can be None
    subject_id : int | str | None
        The id of the subject which the fileset belongs to
    visit_id : int | str | None
        The id of the visit which the fileset belongs to
    repository : Repository
        The repository which the fileset is stored
    from_study : str
        Name of the Arcana study that that generated the field
    exists : bool
        Whether the fileset exists or is just a placeholder for a derivative
    checksums : dict[str, str]
        A checksums of all files within the fileset in a dictionary sorted by
        relative file paths
    record : arcana.pipeline.provenance.Record | None
        The provenance record for the pipeline that generated the file set,
        if applicable
    resource_name : str | None
        For repositories where the name of the file format is saved with the
        data (i.e. XNAT) the name of the resource is to enable straightforward
        format identification
    potential_aux_files : list[str]
        A list of paths to potential files to include in the fileset as
        "side-cars" or headers or in a directory format. Used when the
        format of the fileset is not set when it is detected in the repository
        but determined later from a list of candidates in the specification it
        is matched to.
    quality : str
        The quality label assigned to the fileset (e.g. as is saved on XNAT)
    """

    def __init__(self, name, format=None, frequency='per_session', # @ReservedAssignment @IgnorePep8
                 path=None, aux_files=None, id=None, uri=None, subject_id=None, # @ReservedAssignment @IgnorePep8
                 visit_id=None, repository=None, from_study=None,
                 exists=True, checksums=None, record=None, resource_name=None,
                 potential_aux_files=None, quality=None):
        BaseFileset.__init__(self, name=name, format=format,
                             frequency=frequency)
        BaseItem.__init__(self, subject_id, visit_id, repository,
                          from_study, exists, record)
        if aux_files is not None:
            if path is None:
                raise ArcanaUsageError(
                    "Side cars provided to '{}' fileset ({}) but not primary "
                    "path".format(self.name, aux_files))
            if format is None:
                raise ArcanaUsageError(
                    "Side cars provided to '{}' fileset ({}) but format is "
                    "not specified".format(self.name, aux_files))
        if path is not None:
            path = op.abspath(op.realpath(path))
            if aux_files is None:
                aux_files = {}
            elif set(aux_files.keys()) != set(self.format.aux_files.keys()):
                raise ArcanaUsageError(
                    "Provided side cars for '{}' but expected '{}'"
                    .format("', '".join(aux_files.keys()),
                            "', '".join(self.format.aux_files.keys())))
        self._path = path
        self._aux_files = aux_files if aux_files is not None else {}
        self._uri = uri
        self._id = id
        self._checksums = checksums
        self._resource_name = resource_name
        self._quality = quality
        if potential_aux_files is not None and format is not None:
            raise ArcanaUsageError(
                "Potential paths should only be provided to Fileset.__init__ "
                "({}) when the format of the fileset ({}) is not determined"
                .format(self.name, format))
        if potential_aux_files is not None:
            potential_aux_files = list(potential_aux_files)
        self._potential_aux_files = potential_aux_files

    def __getattr__(self, attr):
        """
        For the convenience of being able to make calls on a fileset that are
        dependent on its format, e.g.

            >>> fileset = Fileset('a_name', format=AnImageFormat())
            >>> fileset.get_header()

        we capture missing attributes and attempt to redirect them to methods
        of the format class that take the fileset as the first argument
        """
        try:
            frmt = self.__dict__['_format']
        except KeyError:
            frmt = None
        else:
            try:
                format_attr = getattr(frmt, attr)
            except AttributeError:
                pass
            else:
                if callable(format_attr):
                    return lambda *args, **kwargs: format_attr(self, *args,
                                                               **kwargs)
        raise AttributeError("Filesets of '{}' format don't have a '{}' "
                             "attribute".format(frmt, attr))

    def __eq__(self, other):
        eq = (BaseFileset.__eq__(self, other) and
              BaseItem.__eq__(self, other) and
              self._aux_files == other._aux_files and
              self._id == other._id and
              self._checksums == other._checksums and
              self._resource_name == other._resource_name and
              self._quality == other._quality)
        # Avoid having to cache fileset in order to test equality unless they
        # are already both cached
        try:
            if self._path is not None and other._path is not None:
                eq &= (self._path == other._path)
        except AttributeError:
            return False
        return eq

    def __hash__(self):
        return (BaseFileset.__hash__(self) ^
                BaseItem.__hash__(self) ^
                hash(self._id) ^
                hash(tuple(sorted(self._aux_files.items()))) ^
                hash(self._checksums) ^
                hash(self._resource_name) ^
                hash(self._quality))

    def __lt__(self, other):
        if isinstance(self.id, int) and isinstance(other.id, basestring):
            return True
        elif isinstance(self.id, basestring) and isinstance(other.id, int):
            return False
        else:
            if self.id == other.id:
                # If ids are equal order depending on study name
                # with acquired (from_study==None) coming first
                if self.from_study is None:
                    return other.from_study is not None
                elif other.from_study is None:
                    return False
                elif self.from_study == other.from_study:
                    if self.format_name is None:
                        return other.format_name is not None
                    elif other.format_name is None:
                        return False
                    else:
                        return self.format_name < other.format_name
                else:
                    return self.from_study < other.from_study
            else:
                return self.id < other.id

    def __repr__(self):
        return ("{}('{}', {}, '{}', subj={}, vis={}, stdy={}{}, exists={}, "
                "quality={}{})"
                .format(
                    type(self).__name__, self.name, self.format,
                    self.frequency, self.subject_id,
                    self.visit_id, self.from_study,
                    (", resource_name='{}'".format(self._resource_name)
                     if self._resource_name is not None else ''),
                    self.exists, self.quality,
                    (", path='{}'".format(self.path)
                     if self._path is not None else '')))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseFileset.find_mismatch(self, other, indent)
        mismatch += BaseItem.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self._path != other._path:
            mismatch += ('\n{}path: self={} v other={}'
                         .format(sub_indent, self._path,
                                 other._path))
        if self._id != other._id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self._id,
                                 other._id))
        if self._checksums != other._checksums:
            mismatch += ('\n{}checksum: self={} v other={}'
                         .format(sub_indent, self.checksums,
                                 other.checksums))
        if self._resource_name != other._resource_name:
            mismatch += ('\n{}format_name: self={} v other={}'
                         .format(sub_indent, self._resource_name,
                                 other._resource_name))
        if self._quality != other._quality:
            mismatch += ('\n{}format_name: self={} v other={}'
                         .format(sub_indent, self._quality,
                                 other._quality))
        return mismatch

    @property
    def path(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.name,
                "Cannot access path of {} as it hasn't been derived yet"
                .format(self))
        if self._path is None:
            if self.repository is not None:
                self.get()  # Retrieve from repository
            else:
                raise ArcanaError(
                    "Neither path nor repository has been set for Fileset("
                    "'{}')".format(self.name))
        return self._path

    def set_path(self, path, aux_files=None):
        if path is not None:
            path = op.abspath(op.realpath(path))
            self._exists = True
        self._path = path
        if aux_files is None:
            self._aux_files = dict(self.format.default_aux_file_paths(path))
        else:
            if set(self.format.aux_files.keys()) != set(aux_files.keys()):
                raise ArcanaUsageError(
                    "Keys of provided side cars ('{}') don't match format "
                    "('{}')".format("', '".join(aux_files.keys()),
                                    "', '".join(self.format.aux_files.keys())))
            self._aux_files = aux_files
        self._checksums = self.calculate_checksums()
        self.put()  # Push to repository

    @path.setter
    def path(self, path):
        self.set_path(path, aux_files=None)

    @property
    def paths(self):
        """Iterates through all files in the set"""
        if self.format is None:
            raise ArcanaFileFormatError(
                "Cannot get paths of fileset ({}) that hasn't had its format "
                "set".format(self))
        if self.format.directory:
            return chain(*((op.join(root, f) for f in files)
                           for root, _, files in os.walk(self.path)))
        else:
            return chain([self.path], self.aux_files.values())

    @property
    def format(self):
        return self._format

    @property
    def quality(self):
        return self._quality

    @format.setter
    def format(self, format):  # @ReservedAssignment
        assert isinstance(format, FileFormat)
        self._format = format
        if format.aux_files and self._path is not None:
            self._aux_files = format.assort_files(
                [self._path] + list(self._potential_aux_files))[1]
        if self._id is None and hasattr(format, 'extract_id'):
            self._id = format.extract_id(self)
        # No longer need to retain potentials after we have assigned the real
        # auxiliaries
        self._potential_aux_files = None

    @property
    def fname(self):
        if self.format is None:
            raise ArcanaFileFormatError(
                "Need to provide format before accessing the filename of {}"
                .format(self))
        return self.name + self.format.ext_str

    @property
    def basename(self):
        return self.name

    @property
    def id(self):
        if self._id is None:
            return self.basename
        else:
            return self._id

    @id.setter
    def id(self, id):  # @ReservedAssignment
        if self._id is None:
            self._id = id
        elif id != self._id:
            raise ArcanaUsageError("Can't change value of ID for {} from {} "
                                   "to {}".format(self, self._id, id))

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        if self._uri is None:
            self._uri = uri
        elif uri != self._uri:
            raise ArcanaUsageError("Can't change value of URI for {} from {} "
                                   "to {}".format(self, self._uri, uri))

    def aux_file(self, name):
        return self._aux_files[name]

    @property
    def aux_files(self):
        return self._aux_files

    @property
    def aux_file_fnames_and_paths(self):
        return ((self.basename + self.format.aux_files[sc_name], sc_path)
                for sc_name, sc_path in self.aux_files.items())

    @property
    def format_name(self):
        if self.format is None:
            name = self._resource_name
        else:
            name = self.format.name
        return name

    @property
    def checksums(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.name,
                "Cannot access checksums of {} as it hasn't been derived yet"
                .format(self))
        if self._checksums is None:
            if self.repository is not None:
                self._checksums = self.repository.get_checksums(self)
            if self._checksums is None:
                self._checksums = self.calculate_checksums()
        return self._checksums

    def calculate_checksums(self):
        checksums = {}
        for fpath in self.paths:
            with open(fpath, 'rb') as f:
                checksums[op.relpath(fpath, self.path)] = hashlib.md5(
                    f.read()).hexdigest()
        return checksums

    @classmethod
    def from_path(cls, path, **kwargs):
        if not op.exists(path):
            raise ArcanaUsageError(
                "Attempting to read Fileset from path '{}' but it "
                "does not exist".format(path))
        if op.isdir(path):
            name = op.basename(path)
        else:
            name = split_extension(op.basename(path))[0]
        return cls(name, path=path, **kwargs)

    def detect_format(self, candidates):
        """
        Detects the format of the fileset from a list of possible
        candidates. If multiple candidates match the potential files, e.g.
        NiFTI-X (see dcm2niix) and NiFTI, then the first matching candidate is
        selected.

        If a 'format_name' was specified when the fileset was
        created then that is used to select between the candidates. Otherwise
        the file extensions of the primary path and potential auxiliary files,
        or extensions of the files within the directory for directories are
        matched against those specified for the file formats

        Parameters
        ----------
        candidates : FileFormat
            A list of file-formats to select from.
        """
        if self._format is not None:
            raise ArcanaFileFormatError(
                "Format has already been set for {}".format(self))
        matches = [c for c in candidates if c.matches(self)]
        if not matches:
            raise ArcanaFileFormatError(
                "None of the candidate file formats ({}) match {}"
                .format(', '.join(str(c) for c in candidates), self))
        return matches[0]

    def initkwargs(self):
        dct = BaseFileset.initkwargs(self)
        dct.update(BaseItem.initkwargs(self))
        dct['path'] = self.path
        dct['id'] = self.id
        dct['uri'] = self.uri
        dct['bids_attr'] = self.bids_attr
        dct['checksums'] = self.checksums
        dct['resource_name'] = self._resource_name
        dct['potential_aux_files'] = self._potential_aux_files
        dct['quality'] = self._quality
        return dct

    def get(self):
        if self.repository is not None:
            self._exists = True
            self._path, self._aux_files = self.repository.get_fileset(self)

    def put(self):
        if self.repository is not None and self._path is not None:
            self.repository.put_fileset(self)

    def contents_equal(self, other, **kwargs):
        """
        Test the equality of the fileset contents with another fileset. If the
        fileset's format implements a 'contents_equal' method than that is used
        to determine the equality, otherwise a straight comparison of the
        checksums is used.

        Parameters
        ----------
        other : Fileset
            The other fileset to compare to
        """
        if hasattr(self.format, 'contents_equal'):
            equal = self.format.contents_equal(self, other, **kwargs)
        else:
            equal = (self.checksums == other.checksums)
        return equal


class Field(BaseItem, BaseField):
    """
    A representation of a value field in the repository.

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    derived : bool
        Whether or not the value belongs in the derived session or not
    subject_id : int | str | None
        The id of the subject which the field belongs to
    visit_id : int | str | None
        The id of the visit which the field belongs to
    repository : Repository
        The repository which the field is stored
    from_study : str
        Name of the Arcana study that that generated the field
    exists : bool
        Whether the field exists or is just a placeholder for a derivative
    record : arcana.pipeline.provenance.Record | None
        The provenance record for the pipeline that generated the field,
        if applicable
    """

    def __init__(self, name, value=None, dtype=None,
                 frequency='per_session', array=None, subject_id=None,
                 visit_id=None, repository=None, from_study=None,
                 exists=True, record=None):
        # Try to determine dtype and array from value if they haven't
        # been provided.
        if value is None:
            if dtype is None:
                raise ArcanaUsageError(
                    "Either 'value' or 'dtype' must be provided to "
                    "Field init")
            array = bool(array)  # Convert to array is None to False
        else:
            value = parse_value(value)
            if isinstance(value, list):
                if array is False:
                    raise ArcanaUsageError(
                        "Array value passed to '{}', which is explicitly not "
                        "an array ({})".format(name, value))
                array = True
            else:
                if array:
                    raise ArcanaUsageError(
                        "Non-array value ({}) passed to '{}', which expects "
                        "array{}".format(value, name,
                                         ('of type {}'.format(dtype)
                                          if dtype is not None else '')))
                array = False
            if dtype is None:
                if array:
                    dtype = type(value[0])
                else:
                    dtype = type(value)
            else:
                # Ensure everything is cast to the correct type
                if array:
                    value = [dtype(v) for v in value]
                else:
                    value = dtype(value)
        BaseField.__init__(self, name, dtype, frequency, array)
        BaseItem.__init__(self, subject_id, visit_id, repository,
                          from_study, exists, record)
        self._value = value

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseItem.__eq__(self, other) and
                self.value == other.value)

    def __hash__(self):
        return (BaseField.__hash__(self) ^
                BaseItem.__hash__(self) ^
                hash(self.value))

    def find_mismatch(self, other, indent=''):
        mismatch = BaseField.find_mismatch(self, other, indent)
        mismatch += BaseItem.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.value != other.value:
            mismatch += ('\n{}value: self={} v other={}'
                         .format(sub_indent, self.value,
                                 other.value))
        return mismatch

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        if self.array:
            val = '[' + ','.join(self._to_str(v) for v in self.value) + ']'
        else:
            val = self._to_str(self.value)
        return val

    def _to_str(self, val):
        if self.dtype is str:
            val = '"{}"'.format(val)
        else:
            val = str(val)
        return val

    def __lt__(self, other):
        if self.name == other.name:
            # If ids are equal order depending on study name
            # with acquired (from_study==None) coming first
            if self.from_study is None:
                return other.from_study is None
            elif other.from_study is None:
                return False
            else:
                return self.from_study < other.from_study
        else:
            return self.name < other.name

    def __repr__(self):
        return ("{}('{}',{} '{}', subj={}, vis={}, stdy={}, exists={})"
                .format(
                    type(self).__name__, self.name,
                    (" {},".format(self._value)
                     if self._value is not None else ''),
                    self.frequency, self.subject_id,
                    self.visit_id, self.from_study,
                    self.exists))

    @property
    def value(self):
        if not self.exists:
            raise ArcanaDataNotDerivedYetError(
                self.name,
                "Cannot access value of {} as it hasn't been "
                "derived yet".format(repr(self)))
        if self._value is None:
            if self.repository is not None:
                self._value = self.repository.get_field(self)
            else:
                raise ArcanaError(
                    "Neither value nor repository has been set for Field("
                    "'{}')".format(self.name))
        return self._value

    @value.setter
    def value(self, value):
        if self.array:
            self._value = [self.dtype(v) for v in value]
        else:
            self._value = self.dtype(value)
        self._exists = True
        self.put()

    @property
    def checksums(self):
        """
        For duck-typing with filesets in checksum management. Instead of a
        checksum, just the value of the field is used
        """
        return self.value

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseItem.initkwargs(self))
        dct['value'] = self.value
        return dct

    def get(self):
        if self.repository is not None:
            self._exists = True
            self._value = self.repository.get_field(self)

    def put(self):
        if self.repository is not None and self._value is not None:
            self.repository.put_field(self)



from past.builtins import basestring
from future.utils import PY3, PY2
import subprocess as sp
import importlib
from itertools import zip_longest
import os.path
import errno
from nipype.interfaces.matlab import MatlabCommand
import shutil
import tempfile
from arcana.exceptions import ArcanaUsageError
if PY2:
    from contextlib2 import ExitStack  # @UnusedImport @UnresolvedImport
    from collections import Iterable  # @UnusedImport @UnresolvedImport
else:
    from contextlib import ExitStack  # @UnusedImport @Reimport
    from collections.abc import Iterable  # @Reimport


PATH_SUFFIX = '_path'
FIELD_SUFFIX = '_field'
CHECKSUM_SUFFIX = '_checksum'

package_dir = os.path.join(os.path.dirname(__file__), '..')

try:
    HOSTNAME = sp.check_output('hostname').strip()
    if PY3:
        HOSTNAME = HOSTNAME.decode('utf-8')
except sp.CalledProcessError:
    HOSTNAME = None


def dir_modtime(dpath):
    """
    Returns the latest modification time of all files/subdirectories in a
    directory
    """
    return max(os.path.getmtime(d) for d, _, _ in os.walk(dpath))


double_exts = ('.tar.gz', '.nii.gz')


def split_extension(path):
    """
    A extension splitter that checks for compound extensions such as
    'file.nii.gz'

    Parameters
    ----------
    filename : str
        A filename to split into base and extension

    Returns
    -------
    base : str
        The base part of the string, i.e. 'file' of 'file.nii.gz'
    ext : str
        The extension part of the string, i.e. 'nii.gz' of 'file.nii.gz'
    """
    for double_ext in double_exts:
        if path.endswith(double_ext):
            return path[:-len(double_ext)], double_ext
    dirname = os.path.dirname(path)
    filename = os.path.basename(path)
    parts = filename.split('.')
    if len(parts) == 1:
        base = filename
        ext = None
    else:
        ext = '.' + parts[-1]
        base = '.'.join(parts[:-1])
    return os.path.join(dirname, base), ext


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


def lower(s):
    if s is None:
        return None
    return s.lower()


if PY3:
    JSON_ENCODING = {'encoding': 'utf-8'}
    from os import makedirs  # @UnusedImport
else:
    JSON_ENCODING = {}

    # Implement makedirs with 'exist_ok' kwarg for Python 2
    def makedirs(path, exist_ok=False, **kwargs):
        try:
            os.makedirs(path, **kwargs)
        except OSError as e:
            if not (exist_ok and e.errno == errno.EEXIST):
                raise


def parse_single_value(value):
    """
    Tries to convert to int, float and then gives up and assumes the value
    is of type string. Useful when excepting values that may be string
    representations of numerical values
    """
    if isinstance(value, basestring):
        try:
            if value.startswith('"') and value.endswith('"'):
                value = str(value[1:-1])
            elif '.' in value:
                value = float(value)
            else:
                value = int(value)
        except ValueError:
            value = str(value)
    elif not isinstance(value, (int, float)):
        raise ArcanaUsageError(
            "Unrecognised type for single value {}".format(value))
    return value


def parse_value(value, dtype=None):
    # Split strings with commas into lists
    if isinstance(value, basestring):
        if value.startswith('[') and value.endswith(']'):
            value = value[1:-1].split(',')
    else:
        # Cast all iterables (except strings) into lists
        try:
            value = list(value)
        except TypeError:
            pass
    if isinstance(value, list):
        if dtype is not None:
            value = [dtype(v) for v in value]
        else:
            value = [parse_single_value(v) for v in value]
        # Check to see if datatypes are consistent
        dtypes = set(type(v) for v in value)
        if len(dtypes) > 1:
            raise ArcanaUsageError(
                "Inconsistent datatypes in values array ({})"
                .format(value))
    elif dtype is not None:
        value = dtype(value)
    else:
        value = parse_single_value(value)
    return value


def run_matlab_cmd(cmd):
    delim = '????????'  # A string that won't occur in the Matlab splash
    matlab_cmd = MatlabCommand(
        script=("fprintf('{}'); fprintf({}); exit;".format(delim, cmd)))
    tmp_dir = tempfile.mkdtemp()
    try:
        result = matlab_cmd.run(cwd=tmp_dir)
        return result.runtime.stdout.split(delim)[1]
    finally:
        shutil.rmtree(tmp_dir)


def iscontainer(*items):
    """
    Checks whether all the provided items are containers (i.e of class list,
    dict, tuple, etc...)
    """
    return all(isinstance(i, Iterable) and not isinstance(i, basestring)
               for i in items)


def find_mismatch(first, second, indent=''):
    """
    Finds where two objects differ, iterating down into nested containers
    (i.e. dicts, lists and tuples) They can be nested containers
    any combination of primary dtypes, str, int, float, dict and lists

    Parameters
    ----------
    first : dict | list | tuple | str | int | float
        The first object to compare
    second : dict | list | tuple | str | int | float
        The other object to compare with the first
    indent : str
        The amount newlines in the output string should be indented. Provide
        the actual indent, i.e. a string of spaces.

    Returns
    -------
    mismatch : str
        Human readable output highlighting where two container differ.
    """

    # Basic case where we are dealing with non-containers
    if not (isinstance(first, type(second)) or
            isinstance(second, type(first))):
        mismatch = (' types: self={} v other={}'
                    .format(type(first).__name__, type(second).__name__))
    elif not iscontainer(first, second):
        mismatch = ': self={} v other={}'.format(first, second)
    else:
        sub_indent = indent + '  '
        mismatch = ''
        if isinstance(first, dict):
            if sorted(first.keys()) != sorted(second.keys()):
                mismatch += (' keys: self={} v other={}'
                             .format(sorted(first.keys()),
                                     sorted(second.keys())))
            else:
                mismatch += ":"
                for k in first:
                    if first[k] != second[k]:
                        mismatch += ("\n{indent}'{}' values{}"
                                     .format(k,
                                             find_mismatch(first[k], second[k],
                                                           indent=sub_indent),
                                             indent=sub_indent))
        else:
            mismatch += ":"
            for i, (f, s) in enumerate(zip_longest(first, second)):
                if f != s:
                    mismatch += ("\n{indent}{} index{}"
                                 .format(i,
                                         find_mismatch(f, s,
                                                       indent=sub_indent),
                                         indent=sub_indent))
    return mismatch


def extract_package_version(package_name):
    version = None
    try:
        module = importlib.import_module(package_name)
    except ImportError:
        if package_name.startswith('py'):
            try:
                module = importlib.import_module(package_name[2:])
            except ImportError:
                pass
    else:
        try:
            version = module.__version__
        except AttributeError:
            pass
    return version


def get_class_info(cls):
    info = {'class': '{}.{}'.format(cls.__module__, cls.__name__)}
    version = extract_package_version(cls.__module__.split('.')[0])
    if version is not None:
        info['pkg_version'] = version
    return info



from .base import (
    split_extension, classproperty, lower, JSON_ENCODING, parse_value,
    run_matlab_cmd, find_mismatch, package_dir, dir_modtime,
    PATH_SUFFIX, FIELD_SUFFIX, CHECKSUM_SUFFIX, ExitStack, makedirs,
    get_class_info, HOSTNAME, extract_package_version)



from copy import copy
from arcana.exceptions import (
    ArcanaError, ArcanaUsageError, ArcanaIndexError)
from .base import BaseFileset, BaseField
from .item import Fileset, Field
from collections import OrderedDict
from operator import itemgetter
from itertools import chain


DICOM_SERIES_NUMBER_TAG = ('0020', '0011')


class BaseCollection(object):
    """
    Base class for collection of filesets and field items
    """

    # For duck-typing with *Spec and *Input objects
    is_spec = False
    skip_missing = False
    drop_if_missing = False
    derivable = False

    def __init__(self, collection, frequency):
        self._frequency = frequency
        if frequency == 'per_study':
            # If wrapped in an iterable
            if not isinstance(collection, self.CollectedClass):
                if len(collection) > 1:
                    raise ArcanaUsageError(
                        "More than one {} passed to {}"
                        .format(self.CONTAINED_CLASS.__name__,
                                type(self).__name__))
                collection = list(collection)
            self._collection = collection
        elif frequency == 'per_session':
            self._collection = OrderedDict()
            for subj_id in sorted(set(c.subject_id for c in collection)):
                self._collection[subj_id] = OrderedDict(
                    sorted(((c.visit_id, c) for c in collection
                            if c.subject_id == subj_id),
                           key=itemgetter(0)))
        elif frequency == 'per_subject':
            self._collection = OrderedDict(
                sorted(((c.subject_id, c) for c in collection),
                       key=itemgetter(0)))
        elif frequency == 'per_visit':
            self._collection = OrderedDict(
                sorted(((c.visit_id, c) for c in collection),
                       key=itemgetter(0)))
        else:
            assert False
        for datum in self:
            if not isinstance(datum, self.CollectedClass):
                raise ArcanaUsageError(
                    "Invalid class {} in {}".format(datum, self))

    def __iter__(self):
        if self._frequency == 'per_study':
            return iter(self._collection)
        elif self._frequency == 'per_session':
            return chain(*(c.values()
                           for c in self._collection.values()))
        else:
            return iter(self._collection.values())

    def __len__(self):
        if self.frequency == 'per_session':
            ln = sum(len(d) for d in self._collection.values())
        else:
            ln = len(self._collection)
        return ln

    def _common_attr(self, collection, attr_name, ignore_none=True):
        attr_set = set(getattr(c, attr_name) for c in collection)
        if ignore_none:
            attr_set -= set([None])
        if len(attr_set) > 1:
            raise ArcanaUsageError(
                "Heterogeneous attributes for '{}' within {}".format(
                    attr_name, self))
        try:
            return next(iter(attr_set))
        except StopIteration:
            return None

    def item(self, subject_id=None, visit_id=None):
        """
        Returns a particular fileset|field in the collection corresponding to
        the given subject and visit_ids. subject_id and visit_id must be
        provided for relevant frequencies. Note that subject_id/visit_id can
        also be provided for non-relevant frequencies, they will just be
        ignored.

        Parameter
        ---------
        subject_id : str
            The subject id of the item to return
        visit_id : str
            The visit id of the item to return
        """

        if self.frequency == 'per_session':
            if subject_id is None or visit_id is None:
                raise ArcanaError(
                    "The 'subject_id' ({}) and 'visit_id' ({}) must be "
                    "provided to get an item from {}".format(
                        subject_id, visit_id, self))
            try:
                subj_dct = self._collection[subject_id]
            except KeyError:
                raise ArcanaIndexError(
                    subject_id,
                    "{} not a subject ID in '{}' collection ({})"
                    .format(subject_id, self.name,
                            ', '.join(self._collection.keys())))
            try:
                fileset = subj_dct[visit_id]
            except KeyError:
                raise ArcanaIndexError(
                    visit_id,
                    "{} not a visit ID in subject {} of '{}' "
                    "collection ({})"
                    .format(visit_id, subject_id, self.name,
                            ', '.join(subj_dct.keys())))
        elif self.frequency == 'per_subject':
            if subject_id is None:
                raise ArcanaError(
                    "The 'subject_id' arg must be provided to get "
                    "the match from {}"
                    .format(self))
            try:
                fileset = self._collection[subject_id]
            except KeyError:
                raise ArcanaIndexError(
                    subject_id,
                    "{} not a subject ID in '{}' collection ({})"
                    .format(subject_id, self.name,
                            ', '.join(self._collection.keys())))
        elif self.frequency == 'per_visit':
            if visit_id is None:
                raise ArcanaError(
                    "The 'visit_id' arg must be provided to get "
                    "the match from {}"
                    .format(self))
            try:
                fileset = self._collection[visit_id]
            except KeyError:
                raise ArcanaIndexError(
                    visit_id,
                    "{} not a visit ID in '{}' collection ({})"
                    .format(visit_id, self.name,
                            ', '.join(self._collection.keys())))
        elif self.frequency == 'per_study':
            try:
                fileset = self._collection[0]
            except IndexError:
                raise ArcanaIndexError(
                    "'{}' Collection is empty so doesn't have a "
                    "per_study node".format(self.name))
        return fileset

    @property
    def collection(self):
        "Used for duck typing Collection objects with Spec and Match "
        "in source and sink initiation"
        return self

    def bind(self, study, **kwargs):  # @UnusedVariable
        """
        Used for duck typing Collection objects with Spec and Match
        in source and sink initiation. Checks IDs match sessions in study.
        """
        if self.frequency == 'per_subject':
            tree_subject_ids = list(study.tree.subject_ids)
            subject_ids = list(self._collection.keys())
            if tree_subject_ids != subject_ids:
                raise ArcanaUsageError(
                    "Subject IDs in collection provided to '{}' ('{}') "
                    "do not match Study tree ('{}')".format(
                        self.name, "', '".join(subject_ids),
                        "', '".join(tree_subject_ids)))
        elif self.frequency == 'per_visit':
            tree_visit_ids = list(study.tree.visit_ids)
            visit_ids = list(self._collection.keys())
            if tree_visit_ids != visit_ids:
                raise ArcanaUsageError(
                    "Subject IDs in collection provided to '{}' ('{}') "
                    "do not match Study tree ('{}')".format(
                        self.name, "', '".join(visit_ids),
                        "', '".join(tree_visit_ids)))
        elif self.frequency == 'per_session':
            for subject in study.tree.subjects:
                if subject.id not in self._collection:
                    raise ArcanaUsageError(
                        "Study subject ID '{}' was not found in colleciton "
                        "provided to '{}' (found '{}')".format(
                            subject.id, self.name,
                            "', '".join(self._collection.keys())))
                for session in subject.sessions:
                    if session.visit_id not in self._collection[subject.id]:
                        raise ArcanaUsageError(
                            "Study visit ID '{}' for subject '{}' was not "
                            "found in colleciton provided to '{}' (found '{}')"
                            .format(subject.id, self.name,
                                    "', '".join(
                                        self._collection[subject.id].keys())))

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name


class FilesetCollection(BaseCollection, BaseFileset):
    """
    A collection of filesets across a study (typically within a repository)

    Parameters
    ----------
    name : str
        Name of the collection
    collection : List[Fileset]
        An iterable of equivalent filesets
    frequency : str
        The frequency of the collection
    format : FileFormat | None
        The file format of the collection (will be determined from filesets
        if not provided).
    """

    CollectedClass = Fileset

    def __init__(self, name, collection, format=None, frequency=None,  # @ReservedAssignment @IgnorePep8
                 candidate_formats=None):
        if format is None and candidate_formats is None:
            raise ArcanaUsageError(
                "Either 'format' or candidate_formats needs to be supplied "
                "during the initialisation of a FilesetCollection ('{}')"
                .format(name))
        collection = list(collection)
        if not collection:
            if format is None:
                format = candidate_formats[0]  # @ReservedAssignment
            if frequency is None:
                raise ArcanaUsageError(
                    "Need to provide explicit frequency for empty "
                    "FilesetCollection")
        else:
            implicit_frequency = self._common_attr(collection,
                                                   'frequency')
            if frequency is None:
                frequency = implicit_frequency
            elif frequency != implicit_frequency:
                raise ArcanaUsageError(
                    "Implicit frequency '{}' does not match explicit "
                    "frequency '{}' for '{}' FilesetCollection"
                    .format(implicit_frequency, frequency, name))
            formatted_collection = []
            for fileset in collection:
                fileset = copy(fileset)
                if fileset.exists and fileset.format is None:
                    fileset.format = (fileset.detect_format(candidate_formats)
                                      if format is None else format)
                formatted_collection.append(fileset)
            collection = formatted_collection
            format = self._common_attr(collection, 'format')  # @ReservedAssignment @IgnorePep8
        BaseFileset.__init__(self, name, format, frequency=frequency)
        BaseCollection.__init__(self, collection, frequency)

    def path(self, subject_id=None, visit_id=None):
        return self.item(
            subject_id=subject_id, visit_id=visit_id).path


class FieldCollection(BaseCollection, BaseField):
    """
    A collection of equivalent filesets (either within a repository)

    Parameters
    ----------
    name : str
        Name of the collection
    collection : List[Fileset]
        An iterable of equivalent filesets
    """

    CollectedClass = Field

    def __init__(self, name, collection, frequency=None, dtype=None,
                 array=None):
        collection = list(collection)
        if collection:
            implicit_frequency = self._common_attr(collection,
                                                   'frequency')
            if frequency is None:
                frequency = implicit_frequency
            elif frequency != implicit_frequency:
                raise ArcanaUsageError(
                    "Implicit frequency '{}' does not match explicit "
                    "frequency '{}' for '{}' FieldCollection"
                    .format(implicit_frequency, frequency, name))
            implicit_dtype = self._common_attr(collection, 'dtype')
            if dtype is None:
                dtype = implicit_dtype  # @ReservedAssignment
            elif dtype != implicit_dtype:
                raise ArcanaUsageError(
                    "Implicit dtype '{}' does not match explicit "
                    "dtype '{}' for '{}' FieldCollection"
                    .format(implicit_dtype, dtype, name))
            implicit_array = self._common_attr(collection, 'array')
            if array is None:
                array = implicit_array
            elif array != implicit_array:
                raise ArcanaUsageError(
                    "Implicit array '{}' does not match explicit "
                    "array '{}' for '{}' FieldCollection"
                    .format(implicit_array, array, name))
        if frequency is None:
            raise ArcanaUsageError(
                "Need to provide explicit frequency for empty "
                "FieldCollection")
        if dtype is None:
            raise ArcanaUsageError(
                "Need to provide explicit dtype for empty "
                "FieldCollection")
        BaseField.__init__(self, name, dtype=dtype, frequency=frequency,
                           array=array)
        BaseCollection.__init__(self, collection, frequency)

    def value(self, subject_id=None, visit_id=None):
        return self.item(subject_id=subject_id, visit_id=visit_id).value



from unittest import TestCase
from arcana.citation import Citation


fsl_cite = Citation(
    short_name="FSL",
    authors=["Smith, S. M.", "Jenkinson, M.",
             "Woolrich, M. W.", "Beckmann, C. F.",
             "Behrens, T. E.", "Johansen- Berg, H.",
             "Bannister, P. R.", "De Luca, M.", "Drobnjak, I.",
             "Flitney, D. E.", "Niazy, R. K.", "Saunders, J.",
             "Vickers, J.", "Zhang, Y.", "De Stefano, N.",
             "Brady, J. M. & Matthews, P. M."],
    title=(
        "Advances in functional and structural MR image "
        "analysis and implementation as FSL"),
    journal="NeuroImage", year=2004, volume=23,
    pages=("S208", "S219"))


mrtrix_cite = Citation(
    short_name="mrtrix",
    authors=["Tournier, J-D"],
    title="MRtrix Package",
    institute="Brain Research Institute, Melbourne, Australia",
    url="https://github.com/MRtrix3/mrtrix3",
    year=2012)


class TestCitation(TestCase):

    def test_citation(self):
        self.assertEqual(fsl_cite, fsl_cite)
        self.assertEqual(fsl_cite, fsl_cite)
        self.assertNotEqual(fsl_cite, mrtrix_cite)



from arcana.utils import ExitStack
from nipype.interfaces.base import (
    traits, DynamicTraitedSpec, Undefined, File, Directory,
    BaseInterface, isdefined)
from itertools import chain
from copy import copy
from arcana.utils import PATH_SUFFIX, FIELD_SUFFIX, CHECKSUM_SUFFIX
from arcana.pipeline.provenance import Record
from arcana.exceptions import ArcanaError, ArcanaDesignError
import logging

logger = logging.getLogger('arcana')

PATH_TRAIT = traits.Either(File(exists=True), Directory(exists=True))
FIELD_TRAIT = traits.Either(traits.Int, traits.Float, traits.Str,
                            traits.List(traits.Int), traits.List(traits.Float),
                            traits.List(traits.Str))
CHECKSUM_TRAIT = traits.Dict(traits.Str(), traits.Str())
# Trait for checksums that may be joined over iterators
JOINED_CHECKSUM_TRAIT = traits.Either(
    CHECKSUM_TRAIT, traits.List(CHECKSUM_TRAIT),
    traits.List(traits.List(CHECKSUM_TRAIT)))


class RepositoryInterface(BaseInterface):
    """
    Parameters
    ----------
    infields : list of str
        Indicates the input fields to be dynamically created

    outfields: list of str
        Indicates output fields to be dynamically created

    See class examples for usage

    """

    def __init__(self, collections):
        super(RepositoryInterface, self).__init__()
        # Protect against iterators
        collections = list(collections)
        # Check for consistent frequencies in collections
        frequencies = set(c.frequency for c in collections)
        if len(frequencies) > 1:
            raise ArcanaError(
                "Attempting to sink multiple frequencies across collections {}"
                .format(', '.join(str(c) for c in collections)))
        elif frequencies:
            # NB: Exclude very rare case where pipeline doesn't have inputs,
            #     would only really happen in unittests
            self._frequency = next(iter(frequencies))
        # Extract set of repositories used to source/sink from/to
        self.repositories = set(chain(*(
            (i.repository for i in c if i.repository is not None)
            for c in collections)))
        # Segregate into fileset and field collections
        self.fileset_collections = [c for c in collections if c.is_fileset]
        self.field_collections = [c for c in collections if c.is_field]

    def __eq__(self, other):
        try:
            return (
                self.fileset_collections == other.fileset_collections and
                self.field_collections == other.field_collections)
        except AttributeError:
            return False

    def __repr__(self):
        return "{}(filesets={}, fields={})".format(
            type(self).__name__, self.fileset_collections,
            self.field_collections)

    def __ne__(self, other):
        return not self == other

    def _run_interface(self, runtime, *args, **kwargs):  # @UnusedVariable
        return runtime

    @property
    def collections(self):
        return chain(self.fileset_collections, self.field_collections)

    @property
    def frequency(self):
        return self._frequency

    @classmethod
    def _add_trait(cls, spec, name, trait_type):
        spec.add_trait(name, trait_type)
        spec.trait_set(trait_change_notify=False, **{name: Undefined})
        # Access the trait (not sure why but this is done in add_traits
        # so I have also done it here
        getattr(spec, name)

    @classmethod
    def field_trait(cls, field):
        if field.array:
            trait = traits.List(field.dtype)
        else:
            trait = field.dtype
        return trait


class RepositorySpec(DynamicTraitedSpec):
    """
    Base class for input and output specifications for repository source
    and sink interfaces
    """
    subject_id = traits.Str(desc="The subject ID")
    visit_id = traits.Str(desc="The visit ID")


class RepositorySourceInputSpec(RepositorySpec):
    """
    Input specification for repository source interfaces.
    """
    prereqs = traits.List(
        desc=("A list of lists of iterator IDs used in prerequisite pipelines."
              " Only passed here to ensure that prerequisites are processed "
              "before this source is run (so that their outputs exist in the "
              "repository)"))


class RepositorySource(RepositoryInterface):
    """
    Parameters
    ----------
    filesets: list
        List of all filesets to be extracted from the repository
    fields: list
        List of all the fields that are to be extracted from the repository
    """

    input_spec = RepositorySourceInputSpec
    output_spec = RepositorySpec
    _always_run = True

    def _outputs(self):
        outputs = super(RepositorySource, self)._outputs()
        # Add traits for filesets to source and their checksums
        for fileset_collection in self.fileset_collections:
            self._add_trait(outputs,
                            fileset_collection.name + PATH_SUFFIX, PATH_TRAIT)
            self._add_trait(outputs,
                            fileset_collection.name + CHECKSUM_SUFFIX,
                            CHECKSUM_TRAIT)
        # Add traits for fields to source
        for field_collection in self.field_collections:
            self._add_trait(outputs,
                            field_collection.name + FIELD_SUFFIX,
                            self.field_trait(field_collection))
        return outputs

    def _list_outputs(self):
        # Directory that holds session-specific
        outputs = self.output_spec().get()
        subject_id = (self.inputs.subject_id
                      if isdefined(self.inputs.subject_id) else None)
        visit_id = (self.inputs.visit_id
                    if isdefined(self.inputs.visit_id) else None)
        outputs['subject_id'] = self.inputs.subject_id
        outputs['visit_id'] = self.inputs.visit_id
        # Source filesets
        with ExitStack() as stack:
            # Connect to set of repositories that the collections come from
            for repository in self.repositories:
                stack.enter_context(repository)
            for fileset_collection in self.fileset_collections:
                fileset = fileset_collection.item(subject_id, visit_id)
                fileset.get()
                outputs[fileset_collection.name + PATH_SUFFIX] = fileset.path
                outputs[fileset_collection.name +
                        CHECKSUM_SUFFIX] = fileset.checksums
            for field_collection in self.field_collections:
                field = field_collection.item(subject_id, visit_id)
                field.get()
                outputs[field_collection.name + FIELD_SUFFIX] = field.value
        return outputs


class RepositorySinkOutputSpec(DynamicTraitedSpec):

    checksums = traits.Either(
        traits.Dict, FIELD_TRAIT,
        desc=("Provenance information sinked with files and fields. Note that"
              "at this stage it is only used as something to connect to the "
              "\"deiterators\" and eventually the \"final\" node after the "
              "pipeline outputs have been sunk"))


class RepositorySink(RepositoryInterface):
    """
    Interface used to sink derivatives into the output repository

    Parameters
    ----------
    collections : *Collection
        The collections of Field and Fileset objects to insert into the
        outputs repositor(y|ies)
    pipeline : arcana.pipeline.Pipeline
        The pipeline that has produced the outputs to sink
    required : list[str]
        Names of derivatives that are required by downstream nodes. Any
        undefined required derivatives that are undefined will raise an error.
    """

    input_spec = RepositorySpec
    output_spec = RepositorySinkOutputSpec

    def __init__(self, collections, pipeline, required=()):
        super(RepositorySink, self).__init__(collections)
        # Add traits for filesets to sink
        for fileset_collection in self.fileset_collections:
            self._add_trait(self.inputs,
                            fileset_collection.name + PATH_SUFFIX,
                            PATH_TRAIT)
        # Add traits for fields to sink
        for field_collection in self.field_collections:
            self._add_trait(self.inputs,
                            field_collection.name + FIELD_SUFFIX,
                            self.field_trait(field_collection))
        # Add traits for checksums/values of pipeline inputs
        self._pipeline_input_filesets = []
        self._pipeline_input_fields = []
        for inpt in pipeline.inputs:
            if inpt.is_fileset:
                trait_t = JOINED_CHECKSUM_TRAIT
            else:
                trait_t = self.field_trait(inpt)
                trait_t = traits.Either(trait_t, traits.List(trait_t),
                                        traits.List(traits.List(trait_t)))
            self._add_trait(self.inputs, inpt.checksum_suffixed_name, trait_t)
            if inpt.is_fileset:
                self._pipeline_input_filesets.append(inpt.name)
            elif inpt.is_field:
                self._pipeline_input_fields.append(inpt.name)
            else:
                assert False
        self._prov = pipeline.prov
        self._pipeline_name = pipeline.name
        self._from_study = pipeline.study.name
        self._required = required

    def _list_outputs(self):
        outputs = self.output_spec().get()
        # Connect iterables (i.e. subject_id and visit_id)
        subject_id = (self.inputs.subject_id
                      if isdefined(self.inputs.subject_id) else None)
        visit_id = (self.inputs.visit_id
                    if isdefined(self.inputs.visit_id) else None)
        missing_inputs = []
        # Collate input checksums into a dictionary
        input_checksums = {n: getattr(self.inputs, n + CHECKSUM_SUFFIX)
                           for n in self._pipeline_input_filesets}
        input_checksums.update({n: getattr(self.inputs, n + FIELD_SUFFIX)
                                for n in self._pipeline_input_fields})
        output_checksums = {}
        with ExitStack() as stack:
            # Connect to set of repositories that the collections come from
            for repository in self.repositories:
                stack.enter_context(repository)
            for fileset_collection in self.fileset_collections:
                fileset = fileset_collection.item(
                    subject_id,
                    visit_id)
                path = getattr(self.inputs,
                               fileset_collection.name + PATH_SUFFIX)
                if not isdefined(path):
                    if fileset.name in self._required:
                        missing_inputs.append(fileset.name)
                    continue  # skip the upload for this fileset
                fileset.path = path  # Push to repository
                output_checksums[fileset.name] = fileset.checksums
            for field_collection in self.field_collections:
                field = field_collection.item(
                    subject_id,
                    visit_id)
                value = getattr(self.inputs,
                                field_collection.name + FIELD_SUFFIX)
                if not isdefined(value):
                    if field.name in self._required:
                        missing_inputs.append(field.name)
                    continue  # skip the upload for this field
                field.value = value  # Push to repository
                output_checksums[field.name] = field.value
            # Add input and output checksums to provenance record and sink to
            # all repositories that have received data (typically only one)
            prov = copy(self._prov)
            prov['inputs'] = input_checksums
            prov['outputs'] = output_checksums
            record = Record(self._pipeline_name, self.frequency, subject_id,
                            visit_id, self._from_study, prov)
            for repository in self.repositories:
                repository.put_record(record)
        if missing_inputs:
            raise ArcanaDesignError(
                "Required derivatives '{}' to were not created by upstream "
                "nodes connected to sink {}".format(
                    "', '".join(missing_inputs), self))
        # Return cache file paths
        outputs['checksums'] = output_checksums
        return outputs



import importlib
from .base import BaseRequirement
from arcana.exceptions import (
    ArcanaRequirementNotFoundError,
    ArcanaVersionNotDetectableError)


class PythonPackageRequirement(BaseRequirement):

    def __init__(self, name, package_name=None, version_attr='__version__',
                 **kwargs):
        super(PythonPackageRequirement, self).__init__(name, **kwargs)
        self._package_name = package_name if package_name is not None else name
        self._version_attr = version_attr

    def __eq__(self, other):
        return (super(PythonPackageRequirement, self).__eq__(other) and
                self._package_name == other._package_name and
                self._version_attr == other._version_attr)

    def __hash__(self):
        return (super().__hash__() ^ hash(self._package_name) ^
                hash(self._version_attr))

    @property
    def package_name(self):
        return self._package_name

    @property
    def version_attr(self):
        return self._version_attr

    def detect_version_str(self):
        try:
            module = importlib.import_module(self.package_name)
        except ImportError:
            raise ArcanaRequirementNotFoundError(
                "Did not find '{}' module/package for {}"
                .format(self.package_name, self))
        if self.version_attr is None:
            raise ArcanaVersionNotDetectableError(
                "Could not detect version of {} as version information is not "
                "provided in package".format(self))
        try:
            version_str = getattr(module, self.version_attr)
        except KeyError:
            raise ArcanaVersionNotDetectableError(
                "Version attribute '{}' is not present in loaded version of {}"
                " ({})".format(self.version_attr, self, self.package_name))
        return version_str



from future.utils import PY3
from .base import BaseRequirement
import subprocess as sp
from arcana.exceptions import (
    ArcanaUsageError, ArcanaRequirementNotFoundError,
    ArcanaVersionNotDetectableError)


class CliRequirement(BaseRequirement):
    """
    Defines a software package that is available on the command line

    Parameters
    ----------
    name : str
        Name of the package
    min_version : tuple(int|str)
        The minimum version required by the node
    max_version : tuple(int|str) | None
        The maximum version that is compatible with the Node
    test_cmd : str
        The name of a command that should be available when the
        requirement is satisfied.
    version_split : function
        A function that splits the version string into major/minor/micro
        parts or equivalent
    references : list[Citation]
        A list of references that should be cited when using this software
        requirement
    """

    def __init__(self, name, test_cmd, version_switch='--version', **kwargs):
        super(CliRequirement, self).__init__(name, **kwargs)
        self._test_cmd = test_cmd
        self._version_switch = version_switch

    def __eq__(self, other):
        return (super(CliRequirement, self).__eq__(other) and
                self._test_cmd == other._test_cmd and
                self._version_switch == other._version_switch)

    def __hash__(self):
        return (super().__hash__() ^ hash(self._test_cmd) ^
                hash(self._version_switch))

    @property
    def test_cmd(self):
        return self._test_cmd

    @property
    def version_switch(self):
        return self._version_switch

    def detect_version_str(self):
        test_cmd_loc = self.locate_command()
        if self.version_switch is None:
            raise ArcanaVersionNotDetectableError(
                "Could not detect version of {} as version information is not "
                "provided by underlying command".format(self))
        try:
            process = sp.Popen((test_cmd_loc, self._version_switch),
                               stdout=sp.PIPE, stderr=sp.PIPE)
            version_str, stderr = process.communicate()
            if not version_str:
                version_str = stderr
        except sp.CalledProcessError as e:
            raise ArcanaUsageError(
                "Problem calling test command ({}) with version switch '{}' "
                "for {}:\n{}".format(
                    test_cmd_loc, self._version_switch, self, e))
        if PY3:
            version_str = version_str.decode('utf-8')
        return version_str

    def locate_command(self, cmd=None):
        if cmd is None:
            cmd = self.test_cmd
        try:
            location = sp.check_output('which {}'.format(cmd), shell=True)
        except sp.CalledProcessError as e:
            if e.returncode == 1:
                raise ArcanaRequirementNotFoundError(
                    "Could not find command {} on system path, which is part "
                    "of {}".format(cmd, self))
            else:
                raise
        if PY3:
            location = location.decode('utf-8')
        location = location.strip()
        return location




__version__ = '0.4.12'

__authors__ = [
    ("Thomas G. Close", "tom.g.close@gmail.com"),
    ("Francesco Sforazzini", "francesco.sforazzini@gmail.com"),
    ("Phillip G. D. Ward", "phillipgdward@gmail.com")]

install_requires = [
    'xnat>=0.3.17',
    'nipype>=1.1.7',
    'pydicom>=1.0.2',
    'networkx>=2.2',
    'fasteners>=0.7.0',
    'future>=0.16.0',
    'pybids>=0.5.1',
    'contextlib2>=0.5.5',
    'deepdiff>=3.3']



from past.builtins import basestring
from builtins import object
from arcana.exceptions import (
    ArcanaMissingDataException, ArcanaNameError)
from arcana.exceptions import ArcanaUsageError
from .base import Study, StudyMetaClass


class MultiStudy(Study):
    """
    Abstract base class for all studies that combine multiple studies
    into a "multi-study".

    Parameters
    ----------
    name : str
        The name of the combined study.
    repository : Repository
        An Repository object that provides access to a DaRIS, XNAT or
        local file system
    processor : Processor
        The processor the processes the derived data when demanded
    inputs : Dict[str, Fileset|Field]
        A dict containing the a mapping between names of study data_specs
        and existing filesets (typically primary from the scanner but can
        also be replacements for generated data_specs)
    parameters : List[Parameter] | Dict[str, (int|float|str)]
        Parameters that are passed to pipelines when they are constructed
        either as a dictionary of key-value pairs or as a list of
        'Parameter' objects. The name and dtype must match ParamSpecs in
        the _param_spec class attribute (see 'add_param_specs').
    subject_ids : List[(int|str)]
        List of subject IDs to restrict the analysis to
    visit_ids : List[(int|str)]
        List of visit IDs to restrict the analysis to
    check_inputs : bool
        Whether to check the inputs to see if any acquired filesets
        are missing
    reprocess : bool
        Whether to reprocess fileset|fields that have been created with
        different parameters and/or pipeline-versions. If False then
        and exception will be thrown if the repository already contains
        matching filesets|fields created with different parameters.

    Class Attrs
    -----------
    add_substudy_specs : list[SubStudySpec]
        Subclasses of MultiStudy typically have a 'add_substudy_specs'
        class member, which defines the sub-studies that make up the
        combined study and the mapping of their fileset names. The key
        of the outer dictionary will be the name of the sub-study, and
        the value is a tuple consisting of the class of the sub-study
        and a map of fileset names from the combined study to the
        sub-study e.g.

            add_substudy_specs = [
                SubStudySpec('t1_study', T1Study, {'magnitude': 't1'}),
                SubStudySpec('t2_study', T2Study, {'magnitude': 't2'})]

            add_data_specs = [
                FilesetSpec('t1', text_format'),
                FilesetSpec('t2', text_format')]
    add_data_specs : List[FilesetSpec|FieldSpec]
        Add's that data specs to the 'data_specs' class attribute,
        which is a dictionary that maps the names of filesets that are
        used and generated by the study to FilesetSpec objects.
    add_param_specs : List[ParamSpec]
        Default parameters for the class
    """

    _substudy_specs = {}

    implicit_cls_attrs = Study.implicit_cls_attrs + ['_substudy_specs']

    def __init__(self, name, repository, processor, inputs,
                 parameters=None, **kwargs):
        try:
            # This works for PY3 as the metaclass inserts it itself if
            # it isn't provided
            metaclass = type(self).__dict__['__metaclass__']
            if not issubclass(metaclass, MultiStudyMetaClass):
                raise KeyError
        except KeyError:
            raise ArcanaUsageError(
                "Need to set MultiStudyMetaClass (or sub-class) as "
                "the metaclass of all classes derived from "
                "MultiStudy")
        super(MultiStudy, self).__init__(
            name, repository, processor, inputs, parameters=parameters,
            **kwargs)
        self._substudies = {}
        for substudy_spec in self.substudy_specs():
            substudy_cls = substudy_spec.study_class
            # Map inputs, data_specs to the substudy
            mapped_inputs = {}
            for data_name in substudy_cls.data_spec_names():
                mapped_name = substudy_spec.map(data_name)
                if mapped_name in self.input_names:
                    mapped_inputs[data_name] = self.input(mapped_name)
                else:
                    try:
                        inpt = self.spec(mapped_name)
                    except ArcanaMissingDataException:
                        pass
                    else:
                        if inpt.derived:
                            mapped_inputs[data_name] = inpt
            # Map parameters to the substudy
            mapped_parameters = {}
            for param_name in substudy_cls.param_spec_names():
                mapped_name = substudy_spec.map(param_name)
                parameter = self._get_parameter(mapped_name)
                mapped_parameters[param_name] = parameter
            # Create sub-study
            substudy = substudy_spec.study_class(
                name + '_' + substudy_spec.name,
                repository, processor, mapped_inputs,
                parameters=mapped_parameters, enforce_inputs=False,
                subject_ids=self.subject_ids, visit_ids=self.visit_ids,
                clear_caches=False)
            # Append to dictionary of substudies
            if substudy_spec.name in self._substudies:
                raise ArcanaNameError(
                    substudy_spec.name,
                    "Duplicate sub-study names '{}'"
                    .format(substudy_spec.name))
            self._substudies[substudy_spec.name] = substudy

    @property
    def substudies(self):
        return iter(self._substudies.values())

    @property
    def substudy_names(self):
        return iter(self._substudies.keys())

    def substudy(self, name):
        try:
            return self._substudies[name]
        except KeyError:
            raise ArcanaNameError(
                name,
                "'{}' not found in sub-studes ('{}')"
                .format(name, "', '".join(self._substudies)))

    @classmethod
    def substudy_spec(cls, name):
        try:
            return cls._substudy_specs[name]
        except KeyError:
            raise ArcanaNameError(
                name,
                "'{}' not found in sub-studes ('{}')"
                .format(name, "', '".join(cls._substudy_specs)))

    @classmethod
    def substudy_specs(cls):
        return iter(cls._substudy_specs.values())

    @classmethod
    def substudy_spec_names(cls):
        return iter(cls._substudy_specs.keys())

    def __repr__(self):
        return "{}(name='{}')".format(
            self.__class__.__name__, self.name)

    @classmethod
    def translate(cls, substudy_name, pipeline_getter, pipeline_arg_names=(),
                  auto_added=False):
        """
        A method for translating pipeline constructors from a sub-study to the
        namespace of a multi-study. Returns a new method that calls the
        sub-study pipeline constructor with appropriate keyword arguments

        Parameters
        ----------
        substudy_name : str
            Name of the sub-study
        pipeline_getter : str
            Name of method used to construct the pipeline in the sub-study
        pipeline_arg_names : tuple[str]
            Names of pipeline arguments passed to the method
        auto_added : bool
            Signify that a method was automatically added by the
            MultiStudyMetaClass. Used in checks when pickling Study
            objects
        """
        assert isinstance(substudy_name, basestring)
        assert isinstance(pipeline_getter, basestring)

        def translated_getter(self, **kwargs):
            substudy_spec = self.substudy_spec(substudy_name)
            pipeline_args = {n: kwargs.pop(n) for n in pipeline_arg_names}
            # Combine mapping of names of sub-study specs with
            return getattr(self.substudy(substudy_name), pipeline_getter)(
                prefix=substudy_name + '_',
                input_map=substudy_spec.name_map,
                output_map=substudy_spec.name_map,
                study=self, name_maps=kwargs,
                **pipeline_args)
        # Add reduce method to allow it to be pickled
        translated_getter.auto_added = auto_added
        return translated_getter


class SubStudySpec(object):
    """
    Specify a study to be included in a MultiStudy class

    Parameters
    ----------
    name : str
        Name for the sub-study
    study_class : type (sub-classed from Study)
        The class of the sub-study
    name_map : dict[str, str]
        A mapping of fileset/field/parameter names from the sub-study
        namespace to the namespace of the MultiStudy. All data-specs
        that are not explicitly mapped are auto-translated using
        the sub-study prefix (name + '_').
    """

    def __init__(self, name, study_class, name_map=None):
        self._name = name
        self._study_class = study_class
        # Fill fileset map with default values before overriding with
        # argument provided to constructor
        self._name_map = name_map if name_map is not None else {}

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return "{}(name='{}', cls={}, name_map={}".format(
            type(self).__name__, self.name, self.study_class,
            self._name_map)

    @property
    def study_class(self):
        return self._study_class

    @property
    def name_map(self):
        nmap = dict((s.name, self.apply_prefix(s.name))
                    for s in self.auto_data_specs)
        nmap.update(self._name_map)
        return nmap

    def map(self, name):
        try:
            return self._name_map[name]
        except KeyError:
            if name not in self.study_class.spec_names():
                raise ArcanaNameError(
                    name,
                    "'{}' doesn't match any filesets, fields, parameters "
                    "in the study class {} ('{}')"
                    .format(name, self.name,
                            self.study_class.__name__,
                            "', '".join(self.study_class.spec_names())))
            return self.apply_prefix(name)

    def apply_prefix(self, name):
        return self.name + '_' + name

    @property
    def auto_data_specs(self):
        """
        Data specs in the sub-study class that are not explicitly provided
        in the name map
        """
        for spec in self.study_class.data_specs():
            if spec.name not in self._name_map:
                yield spec

    @property
    def auto_param_specs(self):
        """
        Parameter pecs in the sub-study class that are not explicitly provided
        in the name map
        """
        for spec in self.study_class.param_specs():
            if spec.name not in self._name_map:
                yield spec


class MultiStudyMetaClass(StudyMetaClass):
    """
    Metaclass for "multi" study classes that automatically adds
    translated data specs and pipelines from sub-study specs if they
    are not explicitly mapped in the spec.
    """

    def __new__(metacls, name, bases, dct):  # @NoSelf @UnusedVariable
        if not any(issubclass(b, MultiStudy) for b in bases):
            raise ArcanaUsageError(
                "MultiStudyMetaClass can only be used for classes that "
                "have MultiStudy as a base class")
        try:
            add_substudy_specs = dct['add_substudy_specs']
        except KeyError:
            add_substudy_specs = dct['add_substudy_specs'] = []
        dct['_substudy_specs'] = substudy_specs = {}
        for base in reversed(bases):
            try:
                substudy_specs.update(
                    (d.name, d) for d in base.substudy_specs())
            except AttributeError:
                pass
        substudy_specs.update(
            (s.name, s) for s in add_substudy_specs)
        if '__metaclass__' not in dct:
            dct['__metaclass__'] = metacls
        cls = StudyMetaClass(name, bases, dct)
        # Loop through all data specs that haven't been explicitly
        # mapped and add a data spec in the multi class.
        for substudy_spec in list(substudy_specs.values()):
            # Map data specs
            for data_spec in substudy_spec.auto_data_specs:
                trans_sname = substudy_spec.apply_prefix(
                    data_spec.name)
                if trans_sname not in cls.data_spec_names():
                    initkwargs = data_spec.initkwargs()
                    initkwargs['name'] = trans_sname
                    if data_spec.derived:
                        trans_pname = substudy_spec.apply_prefix(
                            data_spec.pipeline_getter)
                        initkwargs['pipeline_getter'] = trans_pname
                        # Check to see whether pipeline has already been
                        # translated or always existed in the class (when
                        # overriding default parameters for example)
                        if not hasattr(cls, trans_pname):
                            setattr(cls, trans_pname,
                                    MultiStudy.translate(
                                        substudy_spec.name,
                                        data_spec.pipeline_getter,
                                        pipeline_arg_names=
                                        data_spec.pipeline_arg_names,
                                        auto_added=True))
                    trans_data_spec = type(data_spec)(**initkwargs)
                    # Allow the default input (e.g. an atlas) to translate
                    # any parameter names it needs to use
                    if not data_spec.derived and data_spec.default is not None:
                        try:
                            trans_data_spec.default.translate(substudy_spec)
                        except AttributeError:
                            pass
                    cls._data_specs[trans_sname] = trans_data_spec
            # Map parameter specs
            for param_spec in substudy_spec.auto_param_specs:
                trans_sname = substudy_spec.apply_prefix(
                    param_spec.name)
                if trans_sname not in cls.param_spec_names():
                    renamed_spec = param_spec.renamed(trans_sname)
                    cls._param_specs[
                        renamed_spec.name] = renamed_spec
        # Check all names in name-map correspond to data or parameter
        # specs
        for substudy_spec in list(substudy_specs.values()):
            local_spec_names = list(
                substudy_spec.study_class.spec_names())
            for (local_name,
                 global_name) in substudy_spec._name_map.items():
                if local_name not in local_spec_names:
                    raise ArcanaUsageError(
                        "'{}' in name-map for '{}' sub study spec in {}"
                        "MultiStudy class does not name a spec in {} "
                        "class"
                        .format(local_name, substudy_spec.name,
                                name, substudy_spec.study_class))
                if global_name not in cls.spec_names():
                    raise ArcanaUsageError(
                        "'{}' in name-map for '{}' sub study spec in {}"
                        "MultiStudy class does not name a spec"
                        .format(global_name, substudy_spec.name, name))
        return cls



from past.builtins import basestring
import json
import re
from copy import deepcopy
from pprint import pformat
from datetime import datetime
from deepdiff import DeepDiff
from arcana.exceptions import ArcanaError, ArcanaUsageError
from arcana.__about__ import install_requires


PROVENANCE_VERSION = '1.0'

ARCANA_DEPENDENCIES = [re.split(r'[><=]+', r)[0] for r in install_requires]


class Record(object):
    """
    A representation of the information required to describe the provenance of
    study derivatives. Records the provenance information relevant to a
    specific session, i.e. the general configuration of the pipeline and file
    checksums|field values of the pipeline inputs used to derive the outputs in
    a given session (or visit, subject, study summary). It also records the
    checksums|values of the outputs in order to detect if they have been
    altered outside of Arcana's management (e.g. manual QC/correction)

    Parameters
    ----------
    pipeline_name : str
        Name of the pipeline the record corresponds to
    frequency : str
        The frequency of the record
    subject_id : str | None
        The subject ID the record corresponds to. If None can be a per-visit or
        per-study summary
    visit_id : str | None
        The visit ID the record corresponds to. If None can be a per-subject or
        per-study summary
    from_study : str
        Name of the study that the record was generated by
    prov : dict[str, *]
        A dictionary containing the provenance recorded/to record
    """

    # For duck-typing with Filesets and Fields
    derived = True

    def __init__(self, pipeline_name, frequency, subject_id, visit_id,
                 from_study, prov):
        self._prov = deepcopy(prov)
        self._pipeline_name = pipeline_name
        self._frequency = frequency
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._from_study = from_study
        if 'datetime' not in self._prov:
            self._prov['datetime'] = datetime.now().isoformat()

    def __repr__(self):
        return ("{}(pipeline={}, frequency={}, subject_id={}, visit_id={}, "
                "from_study='{}')".format(
                    type(self).__name__, self.pipeline_name, self.frequency,
                    self.subject_id, self.visit_id, self.from_study))

    def __eq__(self, other):
        return (self._prov == other._prov and
                self._frequency == other._frequency and
                self._subject_id == other._subject_id and
                self._visit_id == other._visit_id and
                self._from_study == other._from_study)

    @property
    def pipeline_name(self):
        return self._pipeline_name

    @property
    def prov(self):
        return self._prov

    @property
    def inputs(self):
        return self._prov['inputs']

    @property
    def outputs(self):
        return self._prov['outputs']

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def visit_id(self):
        return self._visit_id

    @property
    def from_study(self):
        return self._from_study

    @property
    def frequency(self):
        return self._frequency

    @property
    def datetime(self):
        return self._prov['datetime']

    @property
    def provenance_version(self):
        return self._provenance_version

    def save(self, path):
        """
        Saves the provenance object to a JSON file, optionally including
        checksums for inputs and outputs (which are initially produced mid-
        run) to insert during the write

        Parameters
        ----------
        path : str
            Path to save the generated JSON file
        inputs : dict[str, str | list[str] | list[list[str]]] | None
            Checksums of all pipeline inputs used by the pipeline. For inputs
            of matching frequency to the output derivative associated with the
            provenance object, the values of the dictionary will be single
            checksums. If the output is of lower frequency they will be lists
            of checksums or in the case of 'per_session' inputs to 'per_study'
            outputs, lists of lists of checksum. They need to be provided here
            if the provenance object was initialised without checksums
        outputs : dict[str, str] | None
            Checksums of all pipeline outputs. They need to be provided here
            if the provenance object was initialised without checksums
        """
        with open(path, 'w') as f:
            try:
                json.dump(self.prov, f, indent=2)
            except TypeError:
                raise ArcanaError(
                    "Could not serialise provenance record dictionary:\n{}"
                    .format(pformat(self.prov)))

    @classmethod
    def load(cls, pipeline_name, frequency, subject_id, visit_id, from_study,
             path):
        """
        Loads a saved provenance object from a JSON file

        Parameters
        ----------
        path : str
            Path to the provenance file
        frequency : str
            The frequency of the record
        subject_id : str | None
            The subject ID of the provenance record
        visit_id : str | None
            The visit ID of the provenance record
        from_study : str
            Name of the study the derivatives were created for

        Returns
        -------
        record : Record
            The loaded provenance record
        """
        with open(path) as f:
            prov = json.load(f)
        return Record(pipeline_name, frequency, subject_id, visit_id,
                      from_study, prov)

    def mismatches(self, other, include=None, exclude=None):
        """
        Compares information stored within provenance objects with the
        exception of version information to see if they match. Matches are
        constrained to the paths passed to the 'include' kwarg, with the
        exception of sub-paths passed to the 'exclude' kwarg

        Parameters
        ----------
        other : Provenance
            The provenance object to compare against
        include : list[list[str]] | None
            Paths in the provenance to include in the match. If None all are
            incluced
        exclude : list[list[str]] | None
            Paths in the provenance to exclude from the match. In None all are
            excluded
        """
        if include is not None:
            include_res = [self._gen_prov_path_regex(p) for p in include]
        if exclude is not None:
            exclude_res = [self._gen_prov_path_regex(p) for p in exclude]
        diff = DeepDiff(self._prov, other._prov, ignore_order=True)
        # Create regular expresssions for the include and exclude paths in
        # the format that deepdiff uses for nested dictionary/lists

        def include_change(change):
            if include is None:
                included = True
            else:
                included = any(rx.match(change) for rx in include_res)
            if included and exclude is not None:
                included = not any(rx.match(change) for rx in exclude_res)
            return included

        filtered_diff = {}
        for change_type, changes in diff.items():
            if isinstance(changes, dict):
                filtered = dict((k, v) for k, v in changes.items()
                                if include_change(k))
            else:
                filtered = [c for c in changes if include_change(c)]
            if filtered:
                filtered_diff[change_type] = filtered
        return filtered_diff

    @classmethod
    def _gen_prov_path_regex(self, path):
        if isinstance(path, basestring):
            if path.startswith('/'):
                path = path[1:]
            regex = re.compile(r"root\['{}'\].*"
                               .format(r"'\]\['".join(path.split('/'))))
        elif not isinstance(path, re.Pattern):
            raise ArcanaUsageError(
                "Provenance in/exclude paths can either be path strings or "
                "regexes, not '{}'".format(path))
        return regex



from __future__ import division
import logging
from .base import Environment
from arcana.exceptions import (
    ArcanaRequirementNotFoundError, ArcanaVersionNotDetectableError,
    ArcanaVersionError)

logger = logging.getLogger('arcana')


class StaticEnv(Environment):
    """
    Checks to see if requirements are satisfiable by the current
    computing environment. Subclasses can also manage the loading and
    unloading of modules

    Parameters
    ----------
    fail_on_missing : bool
        Raise an error if a requirement is not satisfied
    fail_on_undetectable : bool
        Raise an error if the version of a requirement cannot be detected
    """

    def __init__(self, fail_on_missing=True, fail_on_undetectable=True):
        self._fail_on_missing = fail_on_missing
        self._fail_on_undetectable = fail_on_undetectable
        self._detected_versions = {}

    def satisfy(self, *requirements):
        """
        Checks whether the given requirements are satisfiable within the given
        execution context

        Parameter
        ---------
        requirements : list(Requirement)
            List of requirements to check whether they are satisfiable
        """
        versions = []
        for req_range in requirements:
            try:
                version = self._detected_versions[req_range.name]
            except KeyError:
                try:
                    version = req_range.requirement.detect_version()
                except ArcanaRequirementNotFoundError as e:
                    if self._fail_on_missing:
                        raise
                    else:
                        logger.warning(e)
                except ArcanaVersionNotDetectableError as e:
                    if self._fail_on_undetectable:
                        raise
                    else:
                        logger.warning(e)
                else:
                    self._detected_versions[req_range.name] = version
            if not req_range.within(version):
                raise ArcanaVersionError(
                    "Detected {} version {} is not within requested range {}"
                    .format(req_range.requirement, version, req_range))
            versions.append(version)
        return versions



from past.builtins import basestring
from builtins import object
from abc import ABCMeta
from .file_format import FileFormat
from copy import copy
from logging import getLogger
from arcana.exceptions import ArcanaError
from future.utils import with_metaclass
from future.types import newstr
from arcana.utils import PATH_SUFFIX, FIELD_SUFFIX, CHECKSUM_SUFFIX
logger = getLogger('arcana')


class BaseData(with_metaclass(ABCMeta, object)):

    VALID_FREQUENCIES = ('per_session', 'per_subject', 'per_visit',
                         'per_study')

    is_fileset = False
    is_field = False

    def __init__(self, name, frequency='per_session'):  # @ReservedAssignment @IgnorePep8
        assert name is None or isinstance(name, basestring)
        if frequency not in self.VALID_FREQUENCIES:
            raise ArcanaError(
                "Unrecognised frequency '{}'".format(frequency))
        self._name = name
        self._frequency = frequency

    def __eq__(self, other):
        return (self.basename == other.basename and
                self.frequency == other.frequency)

    def __hash__(self):
        return hash(self.basename) ^ hash(self.frequency)

    def find_mismatch(self, other, indent=''):
        if self != other:
            mismatch = "\n{}{t}('{}') != {t}('{}')".format(
                indent, self.basename, other.basename,
                t=type(self).__name__)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if self.basename != other.basename:
            mismatch += ('\n{}name: self={} v other={}'
                         .format(sub_indent, self.name, other.name))
        if self.frequency != other.frequency:
            mismatch += ('\n{}frequency: self={} v other={}'
                         .format(sub_indent, self.frequency,
                                 other.frequency))
        return mismatch

    def __lt__(self, other):
        return self.name < other.name

    def __ne__(self, other):
        return not (self == other)

    @property
    def name(self):
        return self._name

    @property
    def basename(self):
        return self._name

    @property
    def frequency(self):
        return self._frequency

    def renamed(self, name):
        """
        Duplicate the datum and rename it
        """
        duplicate = copy(self)
        duplicate._name = name
        return duplicate

    def initkwargs(self):
        return {'name': self.name,
                'frequency': self.frequency}


class BaseFileset(with_metaclass(ABCMeta, BaseData)):
    """
    An abstract base class representing either an acquired fileset or the
    specification for a derived fileset.

    Parameters
    ----------
    name : str
        The name of the fileset
    format : FileFormat
        The file format used to store the fileset. Can be one of the
        recognised formats
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    bids_attr : BidsAttr
        A collection of BIDS attributes for the fileset or spec
    """

    is_fileset = True

    def __init__(self, name, format=None, frequency='per_session'):  # @ReservedAssignment @IgnorePep8
        super(BaseFileset, self).__init__(name=name, frequency=frequency)
        assert format is None or isinstance(format, FileFormat)
        self._format = format

    def __eq__(self, other):
        return (super(BaseFileset, self).__eq__(other) and
                self._format == other._format)

    def __hash__(self):
        return (super(BaseFileset, self).__hash__() ^
                hash(self._format))

    def find_mismatch(self, other, indent=''):
        mismatch = super(BaseFileset, self).find_mismatch(other, indent)
        sub_indent = indent + '  '
        if self.format != other.format:
            mismatch += ('\n{}format: self={} v other={}'
                         .format(sub_indent, self.format,
                                 other.format))
        return mismatch

    @property
    def format(self):
        return self._format

    def __repr__(self):
        return ("{}(name='{}', format={}, frequency='{}')"
                .format(self.__class__.__name__, self.name, self.format,
                        self.frequency))

    def initkwargs(self):
        dct = super(BaseFileset, self).initkwargs()
        dct['format'] = self.format
        return dct

    @property
    def suffixed_name(self):
        return self._name + PATH_SUFFIX

    @property
    def checksum_suffixed_name(self):
        return self._name + CHECKSUM_SUFFIX


class BaseField(with_metaclass(ABCMeta, BaseData)):
    """
    An abstract base class representing either an acquired value or the
    specification for a derived value.

    Parameters
    ----------
    name : str
        The name of the fileset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_study',
        specifying whether the fileset is present for each session, subject,
        visit or project.
    array : bool
        Whether the field contains scalar or array data
    """

    is_field = True

    dtypes = (int, float, str)

    def __init__(self, name, dtype, frequency, array=False):
        super(BaseField, self).__init__(name, frequency)
        if dtype not in self.dtypes + (newstr, None):
            raise ArcanaError(
                "Invalid dtype {}, can be one of {}".format(
                    dtype, ', '.join(self._dtype_names())))
        self._dtype = dtype
        self._array = array

    def __eq__(self, other):
        return (super(BaseField, self).__eq__(other) and
                self.dtype == other.dtype and
                self.array == other.array)

    def __hash__(self):
        return (super(BaseField, self).__hash__() ^ hash(self.dtype) ^
                hash(self.array))

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency='{}')"
                .format(self.__class__.__name__, self.name, self.dtype,
                        self.frequency))

    def find_mismatch(self, other, indent=''):
        mismatch = super(BaseField, self).find_mismatch(other, indent)
        sub_indent = indent + '  '
        if self.dtype != other.dtype:
            mismatch += ('\n{}dtype: self={} v other={}'
                         .format(sub_indent, self.dtype,
                                 other.dtype))
        if self.array != other.array:
            mismatch += ('\n{}array: self={} v other={}'
                         .format(sub_indent, self.array,
                                 other.array))
        return mismatch

    @property
    def dtype(self):
        return self._dtype

    @property
    def array(self):
        return self._array

    @classmethod
    def _dtype_names(cls):
        return (d.__name__ for d in cls.dtypes)

    def initkwargs(self):
        dct = super(BaseField, self).initkwargs()
        dct['dtype'] = self.dtype
        dct['array'] = self.array
        return dct

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency='{}', array={})"
                .format(self.__class__.__name__, self.name, self.dtype,
                        self.frequency, self.array))

    @property
    def suffixed_name(self):
        return self._name + FIELD_SUFFIX

    @property
    def checksum_suffixed_name(self):
        return self.suffixed_name



from __future__ import absolute_import
from builtins import range
import os
import os.path as op
import shutil
import json
import time
import unittest
from multiprocessing import Process
from arcana.utils.testing import BaseTestCase
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import IdentityInterface
from arcana.repository.xnat import XnatRepo
from arcana.processor import SingleProc
from arcana.repository.interfaces import RepositorySource, RepositorySink
from arcana.data import InputFilesets
from arcana.utils import PATH_SUFFIX, JSON_ENCODING
from arcana.data.file_format import text_format
from arcana.utils.testing.xnat import (
    TestOnXnatMixin, SERVER, SKIP_ARGS, filter_scans, logger)
from arcana.study import Study, StudyMetaClass
from arcana.data import InputFilesetSpec, FilesetSpec, FieldSpec
from future.utils import with_metaclass


class DummyStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('source1', text_format),
        InputFilesetSpec('source2', text_format, optional=True),
        InputFilesetSpec('source3', text_format, optional=True),
        InputFilesetSpec('source4', text_format, optional=True),
        FilesetSpec('sink1', text_format, 'dummy_pipeline'),
        FilesetSpec('sink3', text_format, 'dummy_pipeline'),
        FilesetSpec('sink4', text_format, 'dummy_pipeline'),
        FilesetSpec('subject_sink', text_format, 'dummy_pipeline',
                    frequency='per_subject'),
        FilesetSpec('visit_sink', text_format, 'dummy_pipeline',
                    frequency='per_visit'),
        FilesetSpec('study_sink', text_format, 'dummy_pipeline',
                    frequency='per_study'),
        FilesetSpec('resink1', text_format, 'dummy_pipeline'),
        FilesetSpec('resink2', text_format, 'dummy_pipeline'),
        FilesetSpec('resink3', text_format, 'dummy_pipeline'),
        FieldSpec('field1', int, 'dummy_pipeline'),
        FieldSpec('field2', float, 'dummy_pipeline'),
        FieldSpec('field3', str, 'dummy_pipeline')]

    def dummy_pipeline(self, **name_maps):
        return self.new_pipeline('dummy_pipeline', name_maps=name_maps)


class TestXnatSourceAndSinkBase(TestOnXnatMixin, BaseTestCase):

    SUBJECT = 'SUBJECT'
    VISIT = 'VISIT'
    STUDY_NAME = 'astudy'
    SUMMARY_STUDY_NAME = 'asummary'

    INPUT_FILESETS = {'source1': 'foo', 'source2': 'bar',
                      'source3': 'wee', 'source4': 'wa'}
    INPUT_FIELDS = {'field1': 1, 'field2': 0.5, 'field3': 'boo'}

    def setUp(self):
        BaseTestCase.setUp(self)
        TestOnXnatMixin.setUp(self)

    def tearDown(self):
        TestOnXnatMixin.tearDown(self)
        BaseTestCase.tearDown(self)


class TestXnatSourceAndSink(TestXnatSourceAndSinkBase):

    @property
    def checksum_sink_project(self):
        return self.project + 'SINK'

    @unittest.skipIf(*SKIP_ARGS)
    def setUp(self):
        TestXnatSourceAndSinkBase.setUp(self)
        self._create_project(self.checksum_sink_project)

    @unittest.skipIf(*SKIP_ARGS)
    def tearDown(self):
        TestXnatSourceAndSinkBase.tearDown(self)
        self._delete_project(self.checksum_sink_project)

    @unittest.skipIf(*SKIP_ARGS)
    def test_repository_roundtrip(self):

        # Create working dirs
        # Create DarisSource node
        repository = XnatRepo(
            project_id=self.project,
            server=SERVER, cache_dir=self.cache_dir)
        study = DummyStudy(
            self.STUDY_NAME, repository, processor=SingleProc('a_dir'),
            inputs=[InputFilesets('source1', 'source1', text_format),
                    InputFilesets('source2', 'source2', text_format),
                    InputFilesets('source3', 'source3', text_format),
                    InputFilesets('source4', 'source4', text_format)])
        # TODO: Should test out other file formats as well.
        source_files = ['source1', 'source2', 'source3', 'source4']
        sink_files = ['sink1', 'sink3', 'sink4']
        inputnode = pe.Node(IdentityInterface(['subject_id',
                                               'visit_id']),
                            'inputnode')
        inputnode.inputs.subject_id = str(self.SUBJECT)
        inputnode.inputs.visit_id = str(self.VISIT)
        source = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in source_files),
            name='source')
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection for f in sink_files),
                dummy_pipeline),
            name='sink')
        sink.inputs.name = 'repository-roundtrip-unittest'
        sink.inputs.desc = (
            "A test session created by repository roundtrip unittest")
        # Create workflow connecting them together
        workflow = pe.Workflow('source-sink-unit-test',
                               base_dir=self.work_dir)
        workflow.add_nodes((source, sink))
        workflow.connect(inputnode, 'subject_id', source, 'subject_id')
        workflow.connect(inputnode, 'visit_id', source, 'visit_id')
        workflow.connect(inputnode, 'subject_id', sink, 'subject_id')
        workflow.connect(inputnode, 'visit_id', sink, 'visit_id')
        for source_name in source_files:
            if source_name != 'source2':
                sink_name = source_name.replace('source', 'sink')
                workflow.connect(
                    source, source_name + PATH_SUFFIX,
                    sink, sink_name + PATH_SUFFIX)
        workflow.run()
        # Check cache was created properly
        self.assertEqual(filter_scans(os.listdir(self.session_cache())),
                         ['source1', 'source2',
                          'source3', 'source4'])
        expected_sink_filesets = ['sink1', 'sink3', 'sink4']
        self.assertEqual(
            filter_scans(os.listdir(self.session_cache(
                from_study=self.STUDY_NAME))), expected_sink_filesets)
        with self._connect() as login:
            fileset_names = filter_scans(login.experiments[self.session_label(
                from_study=self.STUDY_NAME)].scans.keys())
        self.assertEqual(fileset_names, expected_sink_filesets)

    @unittest.skipIf(*SKIP_ARGS)
    def test_fields_roundtrip(self):
        repository = XnatRepo(
            server=SERVER, cache_dir=self.cache_dir,
            project_id=self.project)
        study = DummyStudy(
            self.STUDY_NAME, repository, processor=SingleProc('a_dir'),
            inputs=[InputFilesets('source1', 'source1', text_format)])
        fields = ['field{}'.format(i) for i in range(1, 4)]
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        sink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection for f in fields),
                dummy_pipeline),
            name='fields_sink')
        sink.inputs.field1_field = field1 = 1
        sink.inputs.field2_field = field2 = 2.0
        sink.inputs.field3_field = field3 = str('3')
        sink.inputs.subject_id = self.SUBJECT
        sink.inputs.visit_id = self.VISIT
        sink.inputs.desc = "Test sink of fields"
        sink.inputs.name = 'test_sink'
        sink.run()
        source = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in fields),
            name='fields_source')
        source.inputs.visit_id = self.VISIT
        source.inputs.subject_id = self.SUBJECT
        source.inputs.desc = "Test source of fields"
        source.inputs.name = 'test_source'
        results = source.run()
        self.assertEqual(results.outputs.field1_field, field1)
        self.assertEqual(results.outputs.field2_field, field2)
        self.assertEqual(results.outputs.field3_field, field3)

    @unittest.skip('Skipping delayed download test as it is is proving '
                   'problematic')
    def test_delayed_download(self):
        """
        Tests handling of race conditions where separate processes attempt to
        cache the same fileset
        """
        cache_dir = op.join(self.work_dir,
                                 'cache-delayed-download')
        DATASET_NAME = 'source1'
        target_path = op.join(self.session_cache(cache_dir),
                              DATASET_NAME,
                              DATASET_NAME + text_format.extension)
        tmp_dir = target_path + '.download'
        shutil.rmtree(cache_dir, ignore_errors=True)
        os.makedirs(cache_dir)
        repository = XnatRepo(server=SERVER, cache_dir=cache_dir,
                                    project_id=self.project)
        study = DummyStudy(
            self.STUDY_NAME, repository, SingleProc('ad'),
            inputs=[InputFilesets(DATASET_NAME, DATASET_NAME, text_format)])
        source = pe.Node(
            RepositorySource(
                [study.bound_spec(DATASET_NAME).collection]),
            name='delayed_source')
        source.inputs.subject_id = self.SUBJECT
        source.inputs.visit_id = self.VISIT
        result1 = source.run()
        source1_path = result1.outputs.source1_path
        self.assertTrue(op.exists(source1_path))
        self.assertEqual(source1_path, target_path,
                         "Output file path '{}' not equal to target path '{}'"
                         .format(source1_path, target_path))
        # Clear cache to start again
        shutil.rmtree(cache_dir, ignore_errors=True)
        # Create tmp_dir before running interface, this time should wait for 1
        # second, check to see that the session hasn't been created and then
        # clear it and redownload the fileset.
        os.makedirs(tmp_dir)
        source.inputs.race_cond_delay = 1
        result2 = source.run()
        source1_path = result2.outputs.source1_path
        # Clear cache to start again
        shutil.rmtree(cache_dir, ignore_errors=True)
        # Create tmp_dir before running interface, this time should wait for 1
        # second, check to see that the session hasn't been created and then
        # clear it and redownload the fileset.
        internal_dir = op.join(tmp_dir, 'internal')
        deleted_tmp_dir = tmp_dir + '.deleted'

        def simulate_download():
            "Simulates a download in a separate process"
            os.makedirs(internal_dir)
            time.sleep(5)
            # Modify a file in the temp dir to make the source download keep
            # waiting
            logger.info('Updating simulated download directory')
            with open(op.join(internal_dir, 'download'), 'a') as f:
                f.write('downloading')
            time.sleep(10)
            # Simulate the finalising of the download by copying the previously
            # downloaded file into place and deleting the temp dir.
            logger.info('Finalising simulated download')
            with open(target_path, 'a') as f:
                f.write('simulated')
            shutil.move(tmp_dir, deleted_tmp_dir)

        source.inputs.race_cond_delay = 10
        p = Process(target=simulate_download)
        p.start()  # Start the simulated download in separate process
        time.sleep(1)
        source.run()  # Run the local download
        p.join()
        with open(op.join(deleted_tmp_dir, 'internal', 'download')) as f:
            d = f.read()
        self.assertEqual(d, 'downloading')
        with open(target_path) as f:
            d = f.read()
        self.assertEqual(d, 'simulated')

    @unittest.skipIf(*SKIP_ARGS)
    def test_checksums(self):
        """
        Tests check of downloaded checksums to see if file needs to be
        redownloaded
        """
        cache_dir = op.join(self.work_dir, 'cache-checksum-check')
        DATASET_NAME = 'source1'
        STUDY_NAME = 'checksum_check_study'
        fileset_fname = DATASET_NAME + text_format.extension
        source_target_path = op.join(self.session_cache(cache_dir),
                                     DATASET_NAME)
        md5_path = source_target_path + XnatRepo.MD5_SUFFIX
        source_target_fpath = op.join(source_target_path, fileset_fname)
        shutil.rmtree(cache_dir, ignore_errors=True)
        os.makedirs(cache_dir)
        source_repository = XnatRepo(
            project_id=self.project,
            server=SERVER, cache_dir=cache_dir)
        sink_repository = XnatRepo(
            project_id=self.checksum_sink_project, server=SERVER,
            cache_dir=cache_dir)
        study = DummyStudy(
            STUDY_NAME, sink_repository, SingleProc('ad'),
            inputs=[InputFilesets(DATASET_NAME, DATASET_NAME, text_format,
                                    repository=source_repository)],
            subject_ids=['SUBJECT'], visit_ids=['VISIT'],
            fill_tree=True)
        source = pe.Node(
            RepositorySource(
                [study.bound_spec(DATASET_NAME).collection]),
            name='checksum_check_source')
        source.inputs.subject_id = self.SUBJECT
        source.inputs.visit_id = self.VISIT
        source.run()
        self.assertTrue(op.exists(md5_path))
        self.assertTrue(op.exists(source_target_fpath))
        with open(md5_path) as f:
            checksums = json.load(f)
        # Stash the downloaded file in a new location and create a dummy
        # file instead
        stash_path = source_target_path + '.stash'
        shutil.move(source_target_path, stash_path)
        os.mkdir(source_target_path)
        with open(source_target_fpath, 'w') as f:
            f.write('dummy')
        # Run the download, which shouldn't download as the checksums are the
        # same
        source.run()
        with open(source_target_fpath) as f:
            d = f.read()
        self.assertEqual(d, 'dummy')
        # Replace the checksum with a dummy
        os.remove(md5_path)
        checksums['.'] = 'dummy_checksum'
        with open(md5_path, 'w', **JSON_ENCODING) as f:
            json.dump(checksums, f, indent=2)
        # Retry the download, which should now download since the checksums
        # differ
        source.run()
        with open(source_target_fpath) as f:
            d = f.read()
        with open(op.join(stash_path, fileset_fname)) as f:
            e = f.read()
        self.assertEqual(d, e)
        # Resink the source file and check that the generated MD5 checksum is
        # stored in identical format
        DATASET_NAME = 'sink1'
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        sink = pe.Node(
            RepositorySink(
                [study.bound_spec(DATASET_NAME).collection],
                dummy_pipeline),
            name='checksum_check_sink')
        sink.inputs.name = 'checksum_check_sink'
        sink.inputs.desc = "Tests the generation of MD5 checksums"
        sink.inputs.subject_id = self.SUBJECT
        sink.inputs.visit_id = self.VISIT
        sink.inputs.sink1_path = source_target_fpath
        sink_target_path = op.join(
            self.session_cache(
                cache_dir, project=self.checksum_sink_project,
                subject=(self.SUBJECT), from_study=STUDY_NAME),
            DATASET_NAME)
        sink_md5_path = sink_target_path + XnatRepo.MD5_SUFFIX
        sink.run()
        with open(md5_path) as f:
            source_checksums = json.load(f)
        with open(sink_md5_path) as f:
            sink_checksums = json.load(f)
        self.assertEqual(
            source_checksums, sink_checksums,
            ("Source checksum ({}) did not equal sink checksum ({})"
             .format(source_checksums, sink_checksums)))


class TestXnatSummarySourceAndSink(TestXnatSourceAndSinkBase):

    @unittest.skipIf(*SKIP_ARGS)
    def test_summary(self):
        # Create working dirs
        # Create XnatSource node
        repository = XnatRepo(
            server=SERVER, cache_dir=self.cache_dir,
            project_id=self.project)
        study = DummyStudy(
            self.SUMMARY_STUDY_NAME, repository, SingleProc('ad'),
            inputs=[
                InputFilesets('source1', 'source1', text_format),
                InputFilesets('source2', 'source2', text_format),
                InputFilesets('source3', 'source3', text_format)])
        # TODO: Should test out other file formats as well.
        source_files = ['source1', 'source2', 'source3']
        inputnode = pe.Node(IdentityInterface(['subject_id', 'visit_id']),
                            'inputnode')
        inputnode.inputs.subject_id = self.SUBJECT
        inputnode.inputs.visit_id = self.VISIT
        source = pe.Node(
            RepositorySource(
                [study.bound_spec(f).collection for f in source_files]),
            name='source')
        subject_sink_files = ['subject_sink']
        dummy_pipeline = study.dummy_pipeline()
        dummy_pipeline.cap()
        subject_sink = pe.Node(
            RepositorySink(
                [study.bound_spec(f).collection for f in subject_sink_files],
                dummy_pipeline),
            name='subject_sink')
        subject_sink.inputs.name = 'subject_summary'
        subject_sink.inputs.desc = (
            "Tests the sinking of subject-wide filesets")
        # Test visit sink
        visit_sink_files = ['visit_sink']
        visit_sink = pe.Node(
            RepositorySink(
                [study.bound_spec(f).collection for f in visit_sink_files],
                dummy_pipeline),
            name='visit_sink')
        visit_sink.inputs.name = 'visit_summary'
        visit_sink.inputs.desc = (
            "Tests the sinking of visit-wide filesets")
        # Test project sink
        study_sink_files = ['study_sink']
        study_sink = pe.Node(
            RepositorySink(
                [study.bound_spec(f).collection for f in study_sink_files],
                dummy_pipeline),
            name='study_sink')
        study_sink.inputs.name = 'project_summary'
        study_sink.inputs.desc = (
            "Tests the sinking of project-wide filesets")
        # Create workflow connecting them together
        workflow = pe.Workflow('summary_unittest',
                               base_dir=self.work_dir)
        workflow.add_nodes((source, subject_sink, visit_sink,
                            study_sink))
        workflow.connect(inputnode, 'subject_id', source, 'subject_id')
        workflow.connect(inputnode, 'visit_id', source, 'visit_id')
        workflow.connect(inputnode, 'subject_id', subject_sink, 'subject_id')
        workflow.connect(inputnode, 'visit_id', visit_sink, 'visit_id')
        workflow.connect(
            source, 'source1' + PATH_SUFFIX,
            subject_sink, 'subject_sink' + PATH_SUFFIX)
        workflow.connect(
            source, 'source2' + PATH_SUFFIX,
            visit_sink, 'visit_sink' + PATH_SUFFIX)
        workflow.connect(
            source, 'source3' + PATH_SUFFIX,
            study_sink, 'study_sink' + PATH_SUFFIX)
        workflow.run()
        study.clear_caches()  # Refreshed cached repository tree object
        with self._connect() as login:
            # Check subject summary directories were created properly in cache
            expected_subj_filesets = ['subject_sink']
            subject_dir = self.session_cache(
                visit=XnatRepo.SUMMARY_NAME,
                from_study=self.SUMMARY_STUDY_NAME)
            self.assertEqual(filter_scans(os.listdir(subject_dir)),
                             expected_subj_filesets)
            # and on XNAT
            subject_fileset_names = filter_scans(login.projects[
                self.project].experiments[
                    self.session_label(
                        visit=XnatRepo.SUMMARY_NAME,
                        from_study=self.SUMMARY_STUDY_NAME)].scans.keys())
            self.assertEqual(expected_subj_filesets,
                             subject_fileset_names)
            # Check visit summary directories were created properly in
            # cache
            expected_visit_filesets = ['visit_sink']
            visit_dir = self.session_cache(
                subject=XnatRepo.SUMMARY_NAME,
                from_study=self.SUMMARY_STUDY_NAME)
            self.assertEqual(filter_scans(os.listdir(visit_dir)),
                             expected_visit_filesets)
            # and on XNAT
            visit_fileset_names = filter_scans(login.projects[
                self.project].experiments[
                    self.session_label(
                        subject=XnatRepo.SUMMARY_NAME,
                        from_study=self.SUMMARY_STUDY_NAME)].scans.keys())
            self.assertEqual(expected_visit_filesets, visit_fileset_names)
            # Check project summary directories were created properly in cache
            expected_proj_filesets = ['study_sink']
            project_dir = self.session_cache(
                subject=XnatRepo.SUMMARY_NAME,
                visit=XnatRepo.SUMMARY_NAME,
                from_study=self.SUMMARY_STUDY_NAME)
            self.assertEqual(filter_scans(os.listdir(project_dir)),
                             expected_proj_filesets)
            # and on XNAT
            project_fileset_names = filter_scans(login.projects[
                self.project].experiments[
                    self.session_label(
                        subject=XnatRepo.SUMMARY_NAME,
                        visit=XnatRepo.SUMMARY_NAME,
                        from_study=self.SUMMARY_STUDY_NAME)].scans.keys())
            self.assertEqual(expected_proj_filesets, project_fileset_names)
        # Reload the data from the summary directories
        reloadinputnode = pe.Node(IdentityInterface(['subject_id',
                                                     'visit_id']),
                                  'reload_inputnode')
        reloadinputnode.inputs.subject_id = self.SUBJECT
        reloadinputnode.inputs.visit_id = self.VISIT
        reloadsource_per_subject = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in subject_sink_files),
            name='reload_source_per_subject')
        reloadsource_per_visit = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in visit_sink_files),
            name='reload_source_per_visit')
        reloadsource_per_study = pe.Node(
            RepositorySource(
                study.bound_spec(f).collection for f in study_sink_files),
            name='reload_source_per_study')
        reloadsink = pe.Node(
            RepositorySink(
                (study.bound_spec(f).collection
                 for f in ['resink1', 'resink2', 'resink3']),
                dummy_pipeline),
            name='reload_sink')
        reloadsink.inputs.name = 'reload_summary'
        reloadsink.inputs.desc = (
            "Tests the reloading of subject and project summary filesets")
        reloadworkflow = pe.Workflow('reload_summary_unittest',
                                     base_dir=self.work_dir)
        for node in (reloadsource_per_subject, reloadsource_per_visit,
                     reloadsource_per_study, reloadsink):
            for iterator in ('subject_id', 'visit_id'):
                reloadworkflow.connect(reloadinputnode, iterator,
                                       node, iterator)
        reloadworkflow.connect(reloadsource_per_subject,
                               'subject_sink' + PATH_SUFFIX,
                               reloadsink,
                               'resink1' + PATH_SUFFIX)
        reloadworkflow.connect(reloadsource_per_visit,
                               'visit_sink' + PATH_SUFFIX,
                               reloadsink,
                               'resink2' + PATH_SUFFIX)
        reloadworkflow.connect(reloadsource_per_study,
                               'study_sink' + PATH_SUFFIX,
                               reloadsink,
                               'resink3' + PATH_SUFFIX)
        reloadworkflow.run()
        # Check that the filesets
        self.assertEqual(
            filter_scans(os.listdir(self.session_cache(
                from_study=self.SUMMARY_STUDY_NAME))),
            ['resink1', 'resink2', 'resink3'])
        # and on XNAT
        with self._connect() as login:
            resinked_fileset_names = filter_scans(login.projects[
                self.project].experiments[
                    self.session_label(
                        from_study=self.SUMMARY_STUDY_NAME)].scans.keys())
            self.assertEqual(sorted(resinked_fileset_names),
                             ['resink1', 'resink2', 'resink3'])
